from django.core.urlresolvers import reverse
from django.utils import timezone

from datetime import timedelta

from danceschool.core.constants import REG_VALIDATION_STR, updateConstant
from danceschool.core.utils.tests import DefaultSchoolTestCase
from danceschool.core.models import Invoice, TemporaryRegistration

from .models import PointGroup, PricingTierGroup, DiscountCategory, DiscountCombo, DiscountComboComponent


class BaseDiscountsTest(DefaultSchoolTestCase):

    def create_discount(self,**kwargs):
        '''
        This method just creates the necessary objects to create a simple discount
        with a single required component.
        '''

        test_group, created = PointGroup.objects.get_or_create(name=kwargs.get('pointGroupName','Test points'))
        pt_group, created = PricingTierGroup.objects.get_or_create(
            group=test_group,
            pricingTier=self.defaultPricing,
            points=kwargs.get('pricingTierGroupPoints', 5),
        )

        # Create a flat price combo that just knocks $5 off the regular price
        test_combo = DiscountCombo(
            name=kwargs.get('name','Test Discount'),
            category=kwargs.get('category',DiscountCategory.objects.get(id=1)),
            discountType=kwargs.get('discountType',DiscountCombo.DiscountType.flatPrice),
            onlinePrice=kwargs.get('onlinePrice',self.defaultPricing.onlinePrice - 5),
            doorPrice=kwargs.get('doorPrice',self.defaultPricing.doorPrice - 5),
            dollarDiscount=kwargs.get('dollarDiscount',10),
            percentDiscount=kwargs.get('percentDiscount',50),
            percentUniversallyApplied=kwargs.get('percentUniversallyApplied',False),
            active=kwargs.get('active',True),
            newCustomersOnly=kwargs.get('newCustomersOnly',False),
            daysInAdvanceRequired=kwargs.get('daysInAdvanceRequired',None),
            expirationDate=kwargs.get('expirationDate',None),
        )
        test_combo.save()

        test_component = DiscountComboComponent.objects.create(
            discountCombo=test_combo,
            pointGroup=test_group,
            quantity=kwargs.get('quantity',5),
            allWithinPointGroup=kwargs.get('allWithinPointGroup',False),
        )
        return (test_combo, test_component)

    def register_to_check_discount(self,series):
        '''
        This method makes it easy to determine whether discounts are working
        correctly for a single class registration
        '''

        s = series

        response = self.client.get(reverse('registration'))
        self.assertEqual(response.status_code,200)
        self.assertIn(s, response.context['regOpenSeries'])

        # Sign up for the series, and check that we proceed to the student information page.
        # Because of the way that roles are encoded on this form, we just grab the value to pass
        # from the form itself.
        post_data = {'event_%s' % s.id: response.context_data['form'].fields['event_%s' % s.id].choices[0][0]}

        response = self.client.post(reverse('registration'),post_data,follow=True)
        self.assertEqual(response.redirect_chain,[(reverse('getStudentInfo'), 302)])

        tr = TemporaryRegistration.objects.get(id=self.client.session[REG_VALIDATION_STR].get('temporaryRegistrationId'))
        self.assertTrue(tr.temporaryeventregistration_set.filter(event__id=s.id).exists())

        # Check that the student info page lists the correct item amounts and subtotal
        # with no discounts applied
        self.assertEqual(tr.temporaryeventregistration_set.get(event__id=s.id).price, s.getBasePrice())
        self.assertEqual(response.context_data.get('subtotal'), s.getBasePrice())

        # Continue to the summary page
        post_data = {
            'firstName': 'Discounted',
            'lastName': 'Customer',
            'email': 'test@customer.com',
            'agreeToPolicies': True,
        }
        return self.client.post(reverse('getStudentInfo'),post_data,follow=True)


class DiscountsConditionsTest(BaseDiscountsTest):

    def test_inactive_discount(self):
        '''
        Make a discount inactive and make sure that it doesn't work
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount(active=False)
        s = self.create_series(pricingTier=self.defaultPricing)

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice'))
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),0)
        self.assertFalse(response.context_data.get('addonItems'))
        self.assertFalse(response.context_data.get('discount_codes'))

    def test_expired_discount(self):
        '''
        Create an expired discount and make sure that it doesn't work.
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount(expirationDate=timezone.now() + timedelta(days=-1))
        s = self.create_series(pricingTier=self.defaultPricing)

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice'))
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),0)
        self.assertFalse(response.context_data.get('addonItems'))
        self.assertFalse(response.context_data.get('discount_codes'))

    def test_discounts_disabled(self):
        ''' Disable discounts and check that they don't work anymore '''

        updateConstant('general__discountsEnabled', False)
        test_combo, test_component = self.create_discount()
        s = self.create_series(pricingTier=self.defaultPricing)

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice'))
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),0)
        self.assertFalse(response.context_data.get('addonItems'))
        self.assertFalse(response.context_data.get('discount_codes'))

    def test_notenoughpoints(self):
        '''
        Set the discount's components so that this discount is too small to apply, and
        check that it doesn't get applied.
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount(quantity=10)
        s = self.create_series(pricingTier=self.defaultPricing)

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice'))
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),0)
        self.assertFalse(response.context_data.get('addonItems'))
        self.assertFalse(response.context_data.get('discount_codes'))

    def test_noearlybird(self):
        '''
        Create an early registration discount that requires three day
        advance registration and ensure that it does not work less than
        three days in advance.
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount(daysInAdvanceRequired=3)
        s = self.create_series(pricingTier=self.defaultPricing,startTime=timezone.now() + timedelta(days=1))

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice'))
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),0)
        self.assertFalse(response.context_data.get('addonItems'))
        self.assertFalse(response.context_data.get('discount_codes'))


class DiscountsTypesTest(BaseDiscountsTest):

    def test_discount_applies(self):
        '''
        Create a flat $5 discount and test that it applies
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount()
        s = self.create_series(pricingTier=self.defaultPricing)

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice') - 5)
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),5)
        self.assertFalse(response.context_data.get('addonItems'))

        discount_codes = response.context_data.get('discount_codes')
        self.assertEqual([x[0] for x in discount_codes], [test_combo.name,])

    def test_earlybird(self):
        '''
        Create an early registration discount that requires three day
        advance registration and ensure that it works more than
        three days in advance.
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount(daysInAdvanceRequired=3)
        s = self.create_series(pricingTier=self.defaultPricing,startTime=timezone.now() + timedelta(days=4))

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice') - 5)
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),5)
        self.assertFalse(response.context_data.get('addonItems'))

        discount_codes = response.context_data.get('discount_codes')
        self.assertEqual([x[0] for x in discount_codes], [test_combo.name,])

    def test_allwithinpointgroup(self):
        '''
        Set a discount to apply to an entire point group and check that the price
        is still the flat price
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount(quantity=1, allWithinPointGroup=True)
        s = self.create_series(pricingTier=self.defaultPricing)

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice') - 5)
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),5)
        self.assertFalse(response.context_data.get('addonItems'))

        discount_codes = response.context_data.get('discount_codes')
        self.assertEqual([x[0] for x in discount_codes], [test_combo.name,])

    def test_dollarDiscount(self):
        '''
        Create a $10 off discount and check that it applies appropriately
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount(discountType=DiscountCombo.DiscountType.dollarDiscount, dollarDiscount=10)
        s = self.create_series(pricingTier=self.defaultPricing)

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice') - 10)
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),10)
        self.assertFalse(response.context_data.get('addonItems'))

        discount_codes = response.context_data.get('discount_codes')
        self.assertEqual([x[0] for x in discount_codes], [test_combo.name,])

    def test_percentDiscount(self):
        '''
        Create a 50% off discount and check that it applies correctly.
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount(
            discountType=DiscountCombo.DiscountType.percentDiscount,
            percentDiscount=50,
            percentUniversallyApplied=False
        )
        s = self.create_series(pricingTier=self.defaultPricing)

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),0.5 * response.context_data.get('totalPrice'))
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),0.5 * response.context_data.get('totalPrice'))
        self.assertFalse(response.context_data.get('addonItems'))

        discount_codes = response.context_data.get('discount_codes')
        self.assertEqual([x[0] for x in discount_codes], [test_combo.name,])

    def test_addOnItem(self):
        '''
        Create a free add-on item and ensure that it is applied correctly.
        '''

        updateConstant('general__discountsEnabled', True)
        test_combo, test_component = self.create_discount(
            discountType=DiscountCombo.DiscountType.addOn,
            name='Test Free Add-On',
        )
        s = self.create_series(pricingTier=self.defaultPricing)

        response = self.register_to_check_discount(s)

        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'), response.context_data.get('totalPrice'))
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),0)
        self.assertTrue(response.context_data.get('addonItems'))
        self.assertFalse(response.context_data.get('discount_codes'))

    def test_discountmakesitfree(self):
        '''
        Make the dollar discount larger than the base price and check that
        the registration is free, that the registration is processed and that
        a $0 invoice is created.
        '''

        updateConstant('general__discountsEnabled', True)
        s = self.create_series(pricingTier=self.defaultPricing)
        test_combo, test_component = self.create_discount(discountType=DiscountCombo.DiscountType.dollarDiscount, dollarDiscount=s.getBasePrice() + 10)

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),0)
        self.assertEqual(response.context_data.get('is_free'),True)
        self.assertEqual(response.context_data.get('total_discount_amount'),s.getBasePrice())
        self.assertFalse(response.context_data.get('addonItems'))

        discount_codes = response.context_data.get('discount_codes')
        self.assertEqual([x[0] for x in discount_codes], [test_combo.name,])

        # Since the above registration was free, check that the registration actually
        # processed, and that there exists a paid Invoice for $0
        finalReg = getattr(response.context_data.get('registration'),'registration')
        self.assertTrue(finalReg)
        self.assertEqual(finalReg.netPrice, 0)
        self.assertTrue(finalReg.invoice)
        self.assertTrue(finalReg.invoice.status == Invoice.PaymentStatus.paid)
        self.assertEqual(finalReg.invoice.outstandingBalance, 0)
        self.assertEqual(finalReg.invoice.total, 0)

        # Check that the associated temporary registration is now expired
        self.assertTrue(finalReg.temporaryRegistration.expirationDate <= timezone.now())

        # Show that multiple registrations by the same customer are not permitted
        response = self.register_to_check_discount(s)
        self.assertIn('You are already registered for',' '.join(response.context_data['form'].errors.get('__all__')))

    def test_largerdiscountapplies(self):
        '''
        Create both a $10 discount and a $20 discount, and ensure that the
        larger discount applies
        '''

        updateConstant('general__discountsEnabled', True)
        s = self.create_series(pricingTier=self.defaultPricing)
        test_combo, test_component = self.create_discount(
            discountType=DiscountCombo.DiscountType.dollarDiscount,
            dollarDiscount=10
        )
        bigger_combo, bigger_component = self.create_discount(
            discountType=DiscountCombo.DiscountType.dollarDiscount,
            dollarDiscount=20,
            name='Bigger Discount'
        )

        response = self.register_to_check_discount(s)
        self.assertEqual(response.redirect_chain,[(reverse('showRegSummary'), 302)])
        self.assertEqual(response.context_data.get('totalPrice'), s.getBasePrice())
        self.assertEqual(response.context_data.get('netPrice'),response.context_data.get('totalPrice') - 20)
        self.assertEqual(response.context_data.get('is_free'),False)
        self.assertEqual(response.context_data.get('total_discount_amount'),20)
        self.assertFalse(response.context_data.get('addonItems'))

        discount_codes = response.context_data.get('discount_codes')
        self.assertEqual([x[0] for x in discount_codes], [bigger_combo.name,])



# Third Party Imports
# Give this app a custom verbose name to avoid confusion
from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class BanlistAppConfig(AppConfig):
    name = 'danceschool.banlist'
    verbose_name = _('Registration Ban List Functions')

    def ready(self):
        from . import handlers



from bs4 import BeautifulSoup


def get_text_for_html(html_content):
    '''
    Take the HTML content (from, for example, an email)
    and construct a simple plain text version of that content
    (for example, for inclusion in a multipart email message).
    '''

    soup = BeautifulSoup(html_content)

    # kill all script and style elements
    for script in soup(["script", "style"]):
        script.extract()    # rip it out

    # Replace all links with HREF with the link text and the href in brackets
    for a in soup.findAll('a', href=True):
        a.replaceWith('%s <%s>' % (a.string, a.get('href')))

    # get text
    text = soup.get_text()

    # break into lines and remove leading and trailing space on each
    lines = (line.strip() for line in text.splitlines())
    # break multi-headlines into a line each
    chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
    # drop blank lines
    text = '\n'.join(chunk for chunk in chunks if chunk)

    return text



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-05-12 14:59
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='instructorlistpluginmodel',
            name='title',
            field=models.CharField(blank=True, max_length=200, null=True, verbose_name='Listing Title'),
        ),
        migrations.AlterField(
            model_name='instructorlistpluginmodel',
            name='template',
            field=models.CharField(blank=True, max_length=250, null=True, verbose_name='Template'),
        ),
        migrations.AlterField(
            model_name='locationlistpluginmodel',
            name='template',
            field=models.CharField(blank=True, max_length=250, null=True, verbose_name='Template'),
        ),
        migrations.AlterField(
            model_name='locationpluginmodel',
            name='location',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Location', verbose_name='Location'),
        ),
        migrations.AlterField(
            model_name='locationpluginmodel',
            name='template',
            field=models.CharField(blank=True, max_length=250, null=True, verbose_name='Template'),
        ),
    ]



from django.contrib.auth.views import redirect_to_login
from django.contrib.sites.models import Site
from django.core.exceptions import PermissionDenied, ValidationError
from django.core.urlresolvers import reverse
from django.db.models import Case, When, F, Q, IntegerField, ExpressionWrapper
from django.db.models.functions import ExtractWeekDay
from django.forms import ModelForm, ChoiceField, Media
from django.utils.translation import ugettext_lazy as _
from django.template.loader import get_template, render_to_string
from django.template import Template, Context
from django.template.exceptions import TemplateDoesNotExist


from braces.views import GroupRequiredMixin
from urllib.parse import quote
from six import string_types
import re

from .constants import getConstant
from .tasks import sendEmail
from .registries import plugin_templates_registry


class EmailRecipientMixin(object):

    def email_recipient(self, subject, content, **kwargs):
        '''
        This method allows for direct emailing of an object's recipient(s)
        (default or manually specified), with both object-specific context
        provided using the get_email_context() method.  This is used, for example,
        to email an individual registrant or the recipient of an individual invoice.
        '''

        email_kwargs = {}

        for list_arg in [
            'to','cc','bcc',
        ]:
            email_kwargs[list_arg] = kwargs.pop(list_arg,[]) or []
            if isinstance(email_kwargs[list_arg],string_types):
                email_kwargs[list_arg] = [email_kwargs[list_arg],]

        for none_arg in ['attachment_name','attachment']:
            email_kwargs[none_arg] = kwargs.pop(none_arg,None) or None

        # Ignore any passed HTML content unless explicitly told to send as HTML
        if kwargs.pop('send_html',False) and kwargs.get('html_message'):
            email_kwargs['html_content'] = render_to_string(
                'email/html_email_base.html',
                context={'html_content': kwargs.get('html_message'),'subject': subject}
            )

        email_kwargs['from_name'] = kwargs.pop('from_name',getConstant('email__defaultEmailName')) or \
            getConstant('email__defaultEmailName')
        email_kwargs['from_address'] = kwargs.pop('from_name',getConstant('email__defaultEmailFrom')) or \
            getConstant('email__defaultEmailFrom')

        # Add the object's default recipients if they are provided
        default_recipients = self.get_default_recipients() or []
        if isinstance(default_recipients,string_types):
            default_recipients = [default_recipients,]

        email_kwargs['bcc'] += default_recipients

        if not (email_kwargs['bcc'] or email_kwargs['cc'] or email_kwargs['to']):
            raise ValueError(_('Email must have a recipient.'))

        # In situations where there are no context
        # variables to be rendered, send a mass email
        has_tags = re.search('\{\{.+\}\}',content)
        if not has_tags:
            t = Template(content)
            rendered_content = t.render(Context(kwargs))
            sendEmail(subject,rendered_content,**email_kwargs)
            return

        # Otherwise, get the object-specific email context and email
        # each recipient
        template_context = self.get_email_context() or {}
        template_context.update(kwargs)

        # For security reasons, the following tags are removed from the template before parsing:
        # {% extends %}{% load %}{% debug %}{% include %}{% ssi %}
        content = re.sub(
            '\{%\s*((extends)|(load)|(debug)|(include)|(ssi))\s+.*?\s*%\}',
            '',
            content
        )
        t = Template(content)
        rendered_content = t.render(Context(template_context))

        if email_kwargs.get('html_content'):
            html_content = re.sub(
                '\{%\s*((extends)|(load)|(debug)|(include)|(ssi))\s+.*?\s*%\}',
                '',
                email_kwargs.get('html_content')
            )
            t = Template(html_content)
            email_kwargs['html_content'] = t.render(Context(template_context))

        sendEmail(subject,rendered_content,**email_kwargs)

    def get_email_context(self,**kwargs):
        '''
        This method can be overridden in classes that inherit from this mixin
        so that additional object-specific context is provided to the email
        template.  This should return a dictionary.  By default, only general
        financial context variables are added to the dictionary, and kwargs are
        just passed directly.

        Note also that it is in general not a good idea for security reasons
        to pass model instances in the context here, since these methods can be
        accessed by logged in users who use the SendEmailView.  So, In the default
        models of this app, the values of fields and properties are passed
        directly instead.
        '''
        context = kwargs
        context.update({
            'currencyCode': getConstant('general__currencyCode'),
            'currencySymbol': getConstant('general__currencySymbol'),
            'businessName': getConstant('contact__businessName'),
            'site_url': getConstant('email__linkProtocol') + '://' + Site.objects.get_current().domain,
        })
        return context

    def get_default_recipients(self):
        '''
        This method should be overridden in each class that inherits from this
        mixin, so that the email addresses to whom the email should be sent are
        included on the BCC line.  This should return a list.
        '''
        return []


######################################
# Add general financial variables to the
# context where appropriate.
class FinancialContextMixin(object):
    '''
    This mixin just adds the currency code and symbol to the context
    '''

    def get_context_data(self,**kwargs):
        context = {
            'currencyCode': getConstant('general__currencyCode'),
            'currencySymbol': getConstant('general__currencySymbol'),
            'businessName': getConstant('contact__businessName'),
        }
        context.update(kwargs)
        return super(FinancialContextMixin,self).get_context_data(**context)


class StaffMemberObjectMixin(object):
    '''
    For class-based views involving instructor info, this ensures that
    the object being edited is always for the instructor's own information.
    '''

    def get_object(self, queryset=None):
        if hasattr(self.request.user,'staffmember'):
            return self.request.user.staffmember
        else:
            return None


class GroupRequiredByFieldMixin(GroupRequiredMixin):
    '''
    This subclass of the GroupRequiredMixin checks if a specified model field
    identifies a group that is required.  If so, then require this group.  If
    not, then no permissions are required.  This can be used for thing like survey
    responses that are optionally restricted.
    '''
    group_required_field = ''

    def get_group_required(self):
        ''' Get the group_required value from the object '''
        this_object = self.model_object
        if hasattr(this_object,self.group_required_field):
            if hasattr(getattr(this_object,self.group_required_field),'name'):
                return [getattr(this_object,self.group_required_field).name]
        return ['']

    def check_membership(self, groups):
        ''' Allows for objects with no required groups '''
        if not groups or groups == ['']:
            return True
        if self.request.user.is_superuser:
            return True
        user_groups = self.request.user.groups.values_list("name", flat=True)
        return set(groups).intersection(set(user_groups))

    def dispatch(self, request, *args, **kwargs):
        '''
        This override of dispatch ensures that if no group is required, then
        the request still goes through without being logged in.
        '''
        self.request = request
        in_group = False
        required_group = self.get_group_required()
        if not required_group or required_group == ['']:
            in_group = True
        elif self.request.user.is_authenticated():
            in_group = self.check_membership(required_group)

        if not in_group:
            if self.raise_exception:
                raise PermissionDenied
            else:
                return redirect_to_login(
                    request.get_full_path(),
                    self.get_login_url(),
                    self.get_redirect_field_name())
        return super(GroupRequiredMixin, self).dispatch(
            request, *args, **kwargs)


class AdminSuccessURLMixin(object):
    '''
    Many admin forms should redirect to return to the Page specified in the global settings.
    This mixin is similar to the SuccessURLRedirectListMixin in django-braces.  If a
    success_list_url is specified, it reverses that URL and returns the result.  But,
    if a success_list_url is not specified, then it returns the default admin success form
    Page as specified in settings.
    '''

    success_list_url = None  # Default the success url to none

    def get_success_url(self):
        if self.success_list_url:
            return '%s?redirect_url=%s' % (reverse('submissionRedirect'), quote(self.success_list_url))
        else:
            # If no URL specified, then the redirect view will use the default from the runtime preferences.
            return reverse('submissionRedirect')


class TemplateChoiceField(ChoiceField):

    def validate(self,value):
        '''
        Check for empty values, and for an existing template, but do not check if
        this is one of the initial choices provided.
        '''
        super(ChoiceField,self).validate(value)

        try:
            get_template(value)
        except TemplateDoesNotExist:
            raise ValidationError(_('%s is not a valid template.' % value))


class PluginTemplateMixin(object):
    '''
    This mixin is for plugin classes, to override the render_template with the one provided in the template field,
    and to allow for selectable choices with the option to add a new (validated) choice.
    '''

    def render(self, context, instance, placeholder):
        ''' Permits setting of the template in the plugin instance configuration '''
        if instance and instance.template:
            self.render_template = instance.template
        return super(PluginTemplateMixin,self).render(context,instance,placeholder)

    def templateChoiceFormFactory(self,request,choices):
        class PluginTemplateChoiceForm(ModelForm):

            def __init__(self,*args,**kwargs):
                super(PluginTemplateChoiceForm, self).__init__(*args, **kwargs)

                # Handle passed parameters
                self.request = request
                all_choices = choices

                if self.instance and self.instance.template not in [x[0] for x in choices]:
                    all_choices = [(self.instance.template,self.instance.template)] + all_choices

                if self.request and self.request.user.has_perm('core.choose_custom_plugin_template'):
                    self.fields['template'] = TemplateChoiceField(choices=all_choices)
                else:
                    self.fields['template'] = ChoiceField(choices=all_choices)

            def _media(self):
                ''' Add Select2 custom behavior only if user has permissions to need it. '''
                if self.request and self.request.user.has_perm('core.choose_custom_plugin_template'):
                    return Media(
                        css={'all':('select2/select2.min.css',)},
                        js=('select2/select2.min.js','js/select2_newtemplate.js')
                    )
                return Media()
            media = property(_media)

        return PluginTemplateChoiceForm

    def get_form(self, request, obj=None, **kwargs):
        kwargs['form'] = self.templateChoiceFormFactory(request,self.get_template_choices())
        return super(PluginTemplateMixin, self).get_form(request, obj, **kwargs)

    def get_template_choices(self):
        # If templates are explicitly specified, use those
        if hasattr(self,'template_choices') and self.template_choices:
            return self.template_choices

        # If templates are registered, use those
        registered = [
            x for x in plugin_templates_registry.values() if
            getattr(x,'plugin',None) in [z.__name__ for z in self.__class__.__mro__]
        ]
        if registered:
            return [(x.template_name, getattr(x,'description',None) or x.template_name) for x in registered]

        # If just one template is specified, use that
        if self.render_template:
            return [(self.render_template,self.render_template),]

        # No choices to report
        return []


class EventOrderMixin(object):
    '''
    Various registration pages require that Event querysets be ordered based on the value
    of the constant registration__orgRule (e.g. by session, by month, the combination of the two,
    or by weekday).  Rather than placing this ordering logic in Python, this mixin produces a SQL
    compliant order parameter that can be added as an annotation to any queryset of Events and then
    used to order that queryset.  Since there may be multiple ordering parameters, the function returns
    a tuple.  The first value of the tuple is an ordering parameter that ensures that NULL values
    are always sorted last along both dimensions.  It should always be used in ascending order.
    The second value of the tuple is the primary sort dimension, which can be used ascending or descending
    in the view.  When there is only one ordering parameter needed, the third value of the tuple is None,
    otherwise it passes the second parameter.
    '''

    def get_annotations(self):
        '''
        This method gets the annotations for the queryset.  Unlike get_ordering() below, it
        passes the actual Case() and F() objects that will be evaluated with the queryset, returned
        in a dictionary that is compatible with get_ordering().
        '''
        rule = getConstant('registration__orgRule')

        # Initialize with null values that get filled in based on the logic below.
        annotations = {
            'nullParam': Case(default_value=None,output_field=IntegerField()),
            'paramOne': Case(default_value=None,output_field=IntegerField()),
            'paramTwo': Case(default_value=None,output_field=IntegerField()),
        }

        if rule == 'SessionFirst':
            annotations.update({
                'nullParam': Case(
                    When(session__startTime__isnull=False, then=0),
                    When(month__isnull=False, then=1),
                    default_value=2,
                    output_field=IntegerField()
                ),
                'paramOne': F('session__startTime'),
                'paramTwo': ExpressionWrapper(12 * F('year') + F('month'), output_field=IntegerField()),
            })
        elif rule == 'SessionAlphaFirst':
            annotations.update({
                'nullParam': Case(
                    When(session__name__isnull=False, then=0),
                    When(month__isnull=False, then=1),
                    default_value=2,
                    output_field=IntegerField()
                ),
                'paramOne': F('session__name'),
                'paramTwo': ExpressionWrapper(12 * F('year') + F('month'), output_field=IntegerField()),
            })
        elif rule == 'Month':
            annotations.update({
                'nullParam': Case(
                    When(month__isnull=False, then=0),
                    default_value=1,
                    output_field=IntegerField()
                ),
                'paramOne': ExpressionWrapper(12*F('year') + F('month'), output_field=IntegerField()),
            })
        elif rule == 'Session':
            annotations.update({
                'nullParam': Case(
                    When(session__startTime__isnull=False, then=0),
                    default_value=1,
                    output_field=IntegerField()
                ),
                'paramOne': F('session__startTime'),
            })
        elif rule == 'SessionAlpha':
            annotations.update({
                'nullParam': Case(
                    When(session__name__isnull=False, then=0),
                    default_value=1,
                    output_field=IntegerField()
                ),
                'paramOne': F('session__name'),
            })
        elif rule == 'SessionMonth':
            annotations.update({
                'nullParam': Case(
                    When(Q(session__startTime__isnull=False) & Q(month__isnull=False), then=0),
                    When(Q(session__startTime__isnull=True) & Q(month__isnull=False), then=1),
                    When(Q(session__startTime__isnull=False) & Q(month__isnull=True), then=2),
                    default_value=3,
                    output_field=IntegerField()
                ),
                'paramOne': ExpressionWrapper(12 * F('year') + F('month'), output_field=IntegerField()),
                'paramTwo': F('session__startTime'),
            })
        elif rule == 'SessionAlphaMonth':
            annotations.update({
                'nullParam': Case(
                    When(Q(session__name__isnull=False) & Q(month__isnull=False), then=0),
                    When(Q(session__name__isnull=True) & Q(month__isnull=False), then=1),
                    When(Q(session__name__isnull=False) & Q(month__isnull=True), then=2),
                    default_value=3,
                    output_field=IntegerField()
                ),
                'paramOne': ExpressionWrapper(12*F('year') + F('month'), output_field=IntegerField()),
                'paramTwo': F('session__name'),
            })
        elif rule == 'Weekday':
            annotations.update({
                'nullParam': Case(
                    When(startTime__week_day__isnull=False, then=0),
                    default_value=1,
                    output_field=IntegerField()
                ),
                'paramOne': ExtractWeekDay('startTime'),
            })
        elif rule == 'MonthWeekday':
            annotations.update({
                'nullParam': Case(
                    When(Q(month__isnull=False) & Q(startTime__week_day__isnull=False), then=0),
                    default_value=1,
                    output_field=IntegerField()
                ),
                'paramOne': ExpressionWrapper(12*F('year') + F('month'), output_field=IntegerField()),
                'paramTwo': ExtractWeekDay('startTime'),
            })

        return annotations

    def get_ordering(self, reverseTime=False):
        '''
        This method provides the tuple for ordering of querysets.  However, this will only
        work if the annotations generated by the get_annotations() method above have been
        added to the queryset.  Otherwise, the use of this ordering tuple will fail because
        the appropriate column names will not exist to sort with.
        '''

        # Reverse ordering can be optionally specified in the view class definition.
        reverseTime = getattr(self,'reverse_time_ordering',reverseTime)
        timeParameter = '-startTime' if reverseTime is True else 'startTime'
        return ('nullParam', 'paramOne', 'paramTwo', timeParameter)



# Give this app a custom verbose name to avoid confusion
from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _
from django.core.urlresolvers import reverse


class PrivateEventsAppConfig(AppConfig):
    name = 'danceschool.private_events'
    verbose_name = _('Private Events Functions')

    def ready(self):
        from danceschool.core.models import Location, Room

        @property
        def jsonPrivateCalendarFeed(self):
            '''
            Makes it easy to get location-specific private calendar
            feeds when looping through locations.
            '''
            return reverse('jsonPrivateCalendarFeed', args=(self.id,))
        Location.add_to_class('jsonPrivateCalendarFeed',jsonPrivateCalendarFeed)

        @property
        def jsonRoomPrivateCalendarFeed(self):
            '''
            Makes it easy to get location-specific private calendar
            feeds when looping through locations.
            '''
            return reverse('jsonPrivateCalendarFeed', args=(self.location.id, self.id,))
        Room.add_to_class('jsonPrivateCalendarFeed',jsonRoomPrivateCalendarFeed)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-07-17 16:42
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import djangocms_text_ckeditor.fields


class Migration(migrations.Migration):

    dependencies = [
        ('faq', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='faq',
            options={'ordering': ('orderNum',), 'verbose_name': 'FAQ', 'verbose_name_plural': 'FAQs'},
        ),
        migrations.AlterModelOptions(
            name='faqcategory',
            options={'ordering': ('orderNum',), 'verbose_name': 'FAQ category', 'verbose_name_plural': 'FAQ categories'},
        ),
        migrations.AlterField(
            model_name='faq',
            name='answer',
            field=djangocms_text_ckeditor.fields.HTMLField(help_text='Answer the question.', verbose_name='Answer'),
        ),
        migrations.AlterField(
            model_name='faq',
            name='category',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='faq.FAQCategory', verbose_name='Category'),
        ),
        migrations.AlterField(
            model_name='faq',
            name='creationDate',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Creation date'),
        ),
        migrations.AlterField(
            model_name='faq',
            name='draft',
            field=models.BooleanField(default=False, help_text='Check this box to prevent publication.', verbose_name='Draft status'),
        ),
        migrations.AlterField(
            model_name='faq',
            name='modifiedDate',
            field=models.DateTimeField(auto_now=True, verbose_name='Last modified date'),
        ),
        migrations.AlterField(
            model_name='faq',
            name='orderNum',
            field=models.PositiveIntegerField(default=0, help_text='This number specifies the order in which the questions will be shown on the FAQ page.', verbose_name='Order number'),
        ),
        migrations.AlterField(
            model_name='faq',
            name='question',
            field=models.CharField(max_length=200, verbose_name='Question'),
        ),
        migrations.AlterField(
            model_name='faqcategory',
            name='name',
            field=models.CharField(max_length=200, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='faqcategory',
            name='orderNum',
            field=models.PositiveIntegerField(default=0, help_text='This number specifies the order in which categories will be shown.', verbose_name='Order number'),
        ),
        migrations.AlterField(
            model_name='faqcategorypluginmodel',
            name='category',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='faq.FAQCategory', verbose_name='Category'),
        ),
        migrations.AlterField(
            model_name='faqsinglepluginmodel',
            name='question',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='faq.FAQ', verbose_name='Question'),
        ),
    ]



from huey.contrib.djhuey import db_task


@db_task(retries=3)
def updateSquareFees(paymentRecord):
    '''
    The Square Checkout API does not calculate fees immediately, so this task is
    called to be asynchronously run 1 minute after the initial transaction, so that
    any Invoice or ExpenseItem associated with this transaction also remains accurate.
    '''

    fees = paymentRecord.netFees
    invoice = paymentRecord.invoice
    invoice.fees = fees
    invoice.save()
    invoice.allocateFees()
    return fees



from django.db import models
from django.contrib.auth.models import User, Group
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from colorful.fields import RGBColorField

from danceschool.core.models import Event, EventCategory, EventOccurrence


class PrivateEventCategory(EventCategory):
    '''
    Private events have their own categorization relative to public events,
    though the schema is the same because they both inherit from EventCategory.
    '''
    requiredGroup = models.ForeignKey(Group,verbose_name=_('Group required to add events to this category.'),null=True,blank=True)
    displayColor = RGBColorField(_('Calendar display color'),default='#0000FF')

    class Meta:
        verbose_name = _('Private events category')
        verbose_name_plural = _('Private events categories')


class PrivateEvent(Event):
    '''
    Calendar events may have multiple occurrences.
    '''

    title = models.CharField(_('Title'),max_length=100,help_text=_('Give the event a title'))
    category = models.ForeignKey(PrivateEventCategory,null=True,blank=True,verbose_name=_('Category'))

    @property
    def name(self):
        '''
        Overrides property from Event base class.
        '''
        return self.title

    descriptionField = models.TextField(_('Description'),null=True,blank=True)

    @property
    def description(self):
        '''
        Overrides property from Event base class.
        '''
        return self.descriptionField
    description.fget.short_description = _('Description')

    # Private events can have a location entered other than an officially-listed location
    locationString = models.CharField(_('Other location'),max_length=200,null=True,blank=True,help_text=_('If this event is not at a public event location, then enter it here.'))
    link = models.URLField(_('Optional link'),blank=True,help_text=_('Optionally include the URL to anything that may be relevant for this event.'))

    displayToGroup = models.ForeignKey(Group, null=True, blank=True,verbose_name=_('Display to group'),help_text=_('If this is set, then only these users will see this event on their calendar.'))
    displayToUsers = models.ManyToManyField(User,verbose_name=_('Display to users'),limit_choices_to={'is_staff': True},blank=True,help_text=_('If this is set, then only chosen users will see this event on their calendar.'))

    def __str__(self):
        try:
            return '%s: %s' % (self.name, getattr(self.eventoccurrence_set.first(),'startTime').strftime('%a., %B %d, %Y, %I:%M %p'))
        except AttributeError:
            # Event has no occurrences
            return self.name

    def save(self, *args, **kwargs):
        ''' Set registration status to hidden if it is not specified otherwise '''
        if not self.status:
            self.status == Event.RegStatus.hidden
        super(PrivateEvent,self).save(*args,**kwargs)

    class Meta:
        verbose_name = _('Private event/calendar item')
        verbose_name_plural = _('Private events/calendar items')


class EventReminder(models.Model):
    event = models.ForeignKey(Event,verbose_name=_('Event'))
    eventOccurrence = models.ForeignKey(EventOccurrence,verbose_name=_('Event Occurrence'))

    time = models.DateTimeField(_('Date/Time'))

    notifyList = models.ManyToManyField(User, limit_choices_to={'is_staff': True}, verbose_name=_('Notification List'))
    completed = models.BooleanField(_('Completed'),default=False,help_text=_('This will be set to true once the reminder has been sent.'))

    def save(self,*args,**kwargs):
        if hasattr(self,'event') and not hasattr(self,'eventOccurrence'):
            self.eventOccurrence = self.event.eventoccurrence_set.first()
        if hasattr(self,'eventOccurrence') and not hasattr(self,'event'):
            self.event = self.eventOccurrence.event
        if self.eventOccurrence.event != self.event:
            raise ValidationError(_('Event and EventOccurrence must match!'))
        super(EventReminder,self).save(*args,**kwargs)

    def __str__(self):
        return _('Reminder for ' + self.eventOccurrence.event.name + ': ' + self.time.strftime('%a., %B %d, %Y, %I:%M %p') or '')

    class Meta:
        ordering = ('time',)
        verbose_name = _('Event reminder')
        verbose_name_plural = _('Event reminders')



from django.views.generic import FormView, TemplateView
from django.http import JsonResponse, HttpResponseRedirect, Http404
from django.contrib import messages
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.utils.dateparse import parse_datetime

from datetime import datetime, timedelta

from danceschool.core.models import Instructor, TemporaryRegistration, TemporaryEventRegistration, DanceRole, Event, EventOccurrence, EventStaffMember, Customer
from danceschool.core.constants import getConstant, REG_VALIDATION_STR
from danceschool.core.utils.timezone import ensure_localtime

from .forms import SlotCreationForm, SlotUpdateForm, SlotBookingForm, PrivateLessonStudentInfoForm
from .models import InstructorAvailabilitySlot, PrivateLessonEvent, PrivateLessonCustomer
from .constants import PRIVATELESSON_VALIDATION_STR


class InstructorAvailabilityView(TemplateView):
    template_name = 'private_lessons/instructor_availability_fullcalendar.html'

    def get(self,request,*args,**kwargs):
        # Only instructors or individuals with permission to change
        # other instructors' availability have permission to see this view.
        thisUser = getattr(request,'user',None)
        thisStaffMember = getattr(thisUser,'staffmember',None)
        if (
            (thisStaffMember and thisUser and thisUser.has_perm('private_lessons.edit_own_availability')) or
            (thisUser and thisUser.has_perm('private_lessons.edit_others_availability'))
        ):
            return super(InstructorAvailabilityView,self).get(request,*args,**kwargs)
        raise Http404()

    def get_context_data(self,**kwargs):
        context = super(InstructorAvailabilityView,self).get_context_data(**kwargs)

        context.update({
            'instructor': getattr(getattr(self.request,'user'),'staffmember'),
            'instructor_list': Instructor.objects.filter(
                availableForPrivates=True,instructorprivatelessondetails__isnull=False
            ),
            'creation_form': SlotCreationForm(),
            'update_form': SlotUpdateForm(),
        })

        return context


class AddAvailabilitySlotView(FormView):
    form_class = SlotCreationForm

    def get(self, request, *args, **kwargs):
        return JsonResponse({'valid': False})

    def form_invalid(self, form):
        return JsonResponse(form.errors, status=400)

    def form_valid(self, form):
        '''
        Create slots and return success message.
        '''
        startDate = form.cleaned_data['startDate']
        endDate = form.cleaned_data['endDate']
        startTime = form.cleaned_data['startTime']
        endTime = form.cleaned_data['endTime']
        instructor = form.cleaned_data['instructorId']

        interval_minutes = getConstant('privateLessons__lessonLengthInterval')

        this_date = startDate
        while this_date <= endDate:
            this_time = startTime
            while this_time < endTime:
                InstructorAvailabilitySlot.objects.create(
                    instructor=instructor,
                    startTime=ensure_localtime(datetime.combine(this_date, this_time)),
                    duration=interval_minutes,
                    location=form.cleaned_data.get('location'),
                    room=form.cleaned_data.get('room'),
                    pricingTier=form.cleaned_data.get('pricingTier'),
                )
                this_time = (ensure_localtime(datetime.combine(this_date, this_time)) + timedelta(minutes=interval_minutes)).time()
            this_date += timedelta(days=1)

        return JsonResponse({'valid': True})


class UpdateAvailabilitySlotView(FormView):
    form_class = SlotUpdateForm
    http_method_names = ['post',]

    def get(self, request, *args, **kwargs):
        return JsonResponse({'valid': False})

    def form_invalid(self, form):
        return JsonResponse(form.errors, status=400)

    def form_valid(self, form):
        '''
        Modify or delete the availability slot as requested and return success message.
        '''
        slotIds = form.cleaned_data['slotIds']
        deleteSlot = form.cleaned_data.get('deleteSlot', False)

        these_slots = InstructorAvailabilitySlot.objects.filter(id__in=slotIds)

        if deleteSlot:
            these_slots.delete()
        else:
            for this_slot in these_slots:
                this_slot.location = form.cleaned_data['updateLocation']
                this_slot.room = form.cleaned_data['updateRoom']
                this_slot.status = form.cleaned_data['updateStatus']
                this_slot.pricingTier = form.cleaned_data.get('updatePricing')
                this_slot.save()

        return JsonResponse({'valid': True})


class BookPrivateLessonView(FormView):
    template_name = 'private_lessons/private_lesson_fullcalendar.html'
    form_class = SlotBookingForm

    def get_context_data(self,**kwargs):
        context = super(BookPrivateLessonView,self).get_context_data(**kwargs)
        context.update({
            'instructor_list': Instructor.objects.filter(
                availableForPrivates=True,instructorprivatelessondetails__isnull=False
            ),
            'defaultLessonLength': getConstant('privateLessons__defaultLessonLength'),
        })
        return context

    def get_form_kwargs(self, **kwargs):
        '''
        Pass the current user to the form to render the payAtDoor field if applicable.
        '''
        kwargs = super(BookPrivateLessonView, self).get_form_kwargs(**kwargs)
        kwargs['user'] = self.request.user if hasattr(self.request,'user') else None
        return kwargs

    def form_valid(self, form):

        slotId = form.cleaned_data.pop('slotId')
        payAtDoor = form.cleaned_data.pop('payAtDoor',False)

        # Check that passed duration is valid.
        try:
            duration = int(form.cleaned_data.pop('duration'))
        except ValueError:
            form.add_error(None,ValidationError(_('Invalid duration.'),code='invalid'))
            return self.form_invalid(form)

        # Include the submission user if the user is authenticated
        if self.request.user.is_authenticated:
            submissionUser = self.request.user
        else:
            submissionUser = None

        try:
            thisSlot = InstructorAvailabilitySlot.objects.get(id=slotId)
        except ObjectDoesNotExist:
            form.add_error(None,ValidationError(_('Invalid slot ID.'),code='invalid'))
            return self.form_invalid(form)

        # Check that passed role is valid
        try:
            role = DanceRole.objects.filter(
                instructorprivatelessondetails__instructor=thisSlot.instructor,
            ).get(id=int(form.cleaned_data.pop('role')))
        except (ValueError, ObjectDoesNotExist):
            form.add_error(None,ValidationError(_('Invalid dance role.'),code='invalid'))
            return self.form_invalid(form)

        affectedSlots = InstructorAvailabilitySlot.objects.filter(
            instructor=thisSlot.instructor,
            location=thisSlot.location,
            room=thisSlot.room,
            pricingTier=thisSlot.pricingTier,
            startTime__gte=thisSlot.startTime,
            startTime__lt=thisSlot.startTime + timedelta(minutes=duration),
        ).filter(
            Q(status=InstructorAvailabilitySlot.SlotStatus.available) |
            (
                Q(status=InstructorAvailabilitySlot.SlotStatus.tentative) &
                ~Q(temporaryEventRegistration__registration__expirationDate__gte=timezone.now())
            )
        )

        # If someone cancels, there will already be one or more events associated
        # with these slots.  These need to be deleted, and we also validate to be
        # certain that we are not cancelling any Event which has finalized or in progress
        # registrations attached to it.
        existingEvents = PrivateLessonEvent.objects.filter(
            instructoravailabilityslot__id__in=[x.id for x in affectedSlots]
        ).distinct()

        if existingEvents.filter(
            Q(eventregistration__isnull=False) |
            Q(temporaryeventregistration__registration__expirationDate__gte=timezone.now())
        ).exists():
            form.add_error(None,ValidationError(_('Some or all of your requested lesson time is currently in the process of registration. Please select a new slot or try again later.'),code='invalid'))
            return self.form_invalid(form)
        else:
            existingEvents.delete()

        # Create the lesson record and set related info
        lesson = PrivateLessonEvent.objects.create(
            pricingTier=thisSlot.pricingTier,
            location=thisSlot.location,
            room=thisSlot.room,
            participants=form.cleaned_data.pop('participants'),
            comments=form.cleaned_data.pop('comments'),
            status=Event.RegStatus.hidden,
        )

        lesson_instructor = EventStaffMember.objects.create(
            event=lesson,
            category=getConstant('privateLessons__eventStaffCategoryPrivateLesson'),
            submissionUser=submissionUser,
            staffMember=thisSlot.instructor,
        )

        lesson_occurrence = EventOccurrence.objects.create(
            event=lesson,
            startTime=thisSlot.startTime,
            endTime=thisSlot.startTime + timedelta(minutes=duration),
        )
        lesson_instructor.occurrences.add(lesson_occurrence)

        # Ensure that lesson start and end time are saved appropriately for
        # the event.
        lesson.save()

        # The temporary  expires after a period of inactivity that is specified in preferences.
        expiry = timezone.now() + timedelta(minutes=getConstant('registration__sessionExpiryMinutes'))

        # Slots without pricing tiers can't go through the actual registration process.
        # Instead, they are sent to another view to get contact information.
        if not thisSlot.pricingTier or not getConstant('privateLessons__allowRegistration'):
            affectedSlots.update(
                lessonEvent=lesson,
                status=InstructorAvailabilitySlot.SlotStatus.tentative,
            )
            self.request.session[PRIVATELESSON_VALIDATION_STR] = {
                'lesson': lesson.id,
                'payAtDoor': payAtDoor,
                'expiry': expiry.strftime('%Y-%m-%dT%H:%M:%S%z'),
            }
            return HttpResponseRedirect(reverse('privateLessonStudentInfo'))

        # Slots with pricing tiers require an TemporaryRegistration to be created,
        # and then they are redirected through the registration system.
        else:

            regSession = self.request.session.get(REG_VALIDATION_STR, {})

            # Create a Temporary Registration associated with this lesson.
            reg = TemporaryRegistration(
                submissionUser=submissionUser,dateTime=timezone.now(),
                payAtDoor=payAtDoor,
                expirationDate=expiry,
            )

            tr = TemporaryEventRegistration(
                event=lesson, role=role,
                price=lesson.getBasePrice(payAtDoor=payAtDoor) * affectedSlots.count()
            )

            # Any remaining form data goes into the JSONfield.
            reg.data = form.cleaned_data or {}

            # Now we are ready to save and proceed.
            reg.priceWithDiscount = tr.price
            reg.save()
            tr.registration = reg
            tr.save()

            affectedSlots.update(
                lessonEvent=lesson,
                status=InstructorAvailabilitySlot.SlotStatus.tentative,
                temporaryEventRegistration=tr,
            )

            # Load the temporary registration into session data like a regular registration
            # and redirect to Step 2 as usual.
            regSession["temporaryRegistrationId"] = reg.id
            regSession["temporaryRegistrationExpiry"] = expiry.strftime('%Y-%m-%dT%H:%M:%S%z')
            self.request.session[REG_VALIDATION_STR] = regSession
            return HttpResponseRedirect(reverse('getStudentInfo'))


class PrivateLessonStudentInfoView(FormView):
    '''
    For private lessons booked and paid for using the traditional
    online registration system, that system collects all information
    needed for booking.  However, for lessons booked without the option
    of online payment, we still need to collect the customer's name and
    email adress before finalizing booking.  This view handles the
    collection of that information.
    '''
    template_name = 'private_lessons/get_student_info.html'
    form_class = PrivateLessonStudentInfoForm

    def dispatch(self,request,*args,**kwargs):
        '''
        Handle the session data passed by the prior view.
        '''

        lessonSession = request.session.get(PRIVATELESSON_VALIDATION_STR,{})

        try:
            self.lesson = PrivateLessonEvent.objects.get(id=lessonSession.get('lesson'))
        except (ValueError, ObjectDoesNotExist):
            messages.error(request,_('Invalid lesson identifier passed to sign-up form.'))
            return HttpResponseRedirect(reverse('bookPrivateLesson'))

        expiry = parse_datetime(lessonSession.get('expiry',''),)
        if not expiry or expiry < timezone.now():
            messages.info(request,_('Your registration session has expired. Please try again.'))
            return HttpResponseRedirect(reverse('bookPrivateLesson'))

        self.payAtDoor = lessonSession.get('payAtDoor',False)
        return super(PrivateLessonStudentInfoView,self).dispatch(request,*args,**kwargs)

    def get_context_data(self,**kwargs):
        context = super(PrivateLessonStudentInfoView,self).get_context_data(**kwargs)
        context.update({
            'lesson': self.lesson,
            'teachers': [x.staffMember.fullName for x in self.lesson.eventstaffmember_set.all()],
        })
        return context

    def get_form_kwargs(self, **kwargs):
        ''' Pass along the request data to the form '''
        kwargs = super(PrivateLessonStudentInfoView, self).get_form_kwargs(**kwargs)
        kwargs['request'] = self.request
        kwargs['payAtDoor'] = self.payAtDoor
        return kwargs

    def form_valid(self,form):
        first_name = form.cleaned_data.get('firstName')
        last_name = form.cleaned_data.get('lastName')
        email = form.cleaned_data.get('email')
        phone = form.cleaned_data.get('phone')

        customer, created = Customer.objects.update_or_create(
            first_name=first_name,last_name=last_name,email=email,defaults={'phone': phone}
        )
        # Ensure that this customer is affiliated with this lesson.
        PrivateLessonCustomer.objects.get_or_create(
            customer=customer,
            lesson=self.lesson,
        )

        self.lesson.finalizeBooking()
        messages.success(self.request,_('Your private lesson has been scheduled successfully.'))
        self.request.session.pop(PRIVATELESSON_VALIDATION_STR,{})
        return HttpResponseRedirect(reverse('submissionRedirect'))



from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from cms.toolbar_pool import toolbar_pool
from cms.toolbar_base import CMSToolbar


@toolbar_pool.register
class RequirementLinksToolbar(CMSToolbar):
    ''' Add link to Requirements admin to the Events Related Items '''

    def populate(self):
        if not self.request.user.has_perm('prerequisites.change_requirement'):
            return

        menu = self.toolbar.get_or_create_menu('core-events', _('Events'))
        related_menu = menu.get_or_create_menu('core-events-related',_('Related Items'))
        related_menu.add_link_item(_('Registration Requirements/Prerequisites'), url=reverse('admin:prerequisites_requirement_changelist'))



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-04-25 00:10
from __future__ import unicode_literals

import cms.models.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('cms', '0016_auto_20160608_1535'),
        ('financial', '0001_initial'),
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('invoiceNumber', models.CharField(max_length=100, unique=True, verbose_name='Internal Invoice Number')),
                ('paypalInvoiceID', models.CharField(max_length=100, unique=True, verbose_name='Paypal Invoice ID')),
                ('creationDate', models.DateTimeField(auto_now_add=True)),
                ('paymentDate', models.DateTimeField(blank=True, null=True)),
                ('invoiceURL', models.URLField(blank=True, null=True, verbose_name='Invoice View URL')),
                ('payerViewURL', models.URLField(blank=True, null=True, verbose_name='Invoice Payment URL')),
                ('totalAmount', models.FloatField(verbose_name='Invoice Total Amount (net of discounts)')),
                ('itemList', models.TextField(blank=True, null=True, verbose_name='Item List Passed to Invoice')),
            ],
        ),
        migrations.CreateModel(
            name='IPNCartItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('invoiceName', models.CharField(blank=True, max_length=100, null=True, verbose_name='Invoice Item Name')),
                ('invoiceNumber', models.CharField(max_length=50, verbose_name='Invoice Item Number')),
                ('mc_gross', models.FloatField(default=0, verbose_name='Gross Amount')),
                ('refundAmount', models.FloatField(default=0, verbose_name='Allocated Refund Amount')),
            ],
            options={
                'verbose_name': 'Paypal IPN Cart Item',
            },
        ),
        migrations.CreateModel(
            name='IPNMessage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.TextField(blank=True, null=True)),
                ('invoice', models.CharField(max_length=80, verbose_name='Invoice number')),
                ('generated_invoice', models.CharField(blank=True, max_length=100, null=True, verbose_name='Paypal-generated invoice number')),
                ('mc_fee', models.FloatField()),
                ('mc_gross', models.FloatField()),
                ('payment_status', models.CharField(max_length=30)),
                ('payment_date', models.DateTimeField()),
                ('txn_id', models.CharField(max_length=30, unique=True)),
                ('txn_type', models.CharField(max_length=30)),
                ('custom', models.TextField()),
                ('mc_currency', models.CharField(max_length=30)),
                ('payer_email', models.EmailField(max_length=254)),
                ('receiver_email', models.EmailField(max_length=254)),
                ('payer_id', models.CharField(max_length=30)),
                ('finalRegistration', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Registration', verbose_name='Final Registration')),
                ('paypalInvoice', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='paypal.Invoice', verbose_name='Paypal-Submitted Invoice')),
                ('priorTransaction', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='subsequenttransactions', to='paypal.IPNMessage')),
                ('registration', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='core.TemporaryRegistration', verbose_name='Initial Temporary Registration')),
            ],
            options={
                'permissions': (('allocate_refunds', 'Can allocate Paypal refunds to individual registration items'),),
                'verbose_name': 'Paypal IPN message',
            },
        ),
        migrations.CreateModel(
            name='PayNowFormModel',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='paypal_paynowformmodel', serialize=False, to='cms.CMSPlugin')),
                ('defaultAmount', models.FloatField(default=0, help_text='The initial value for gift certificate forms.', verbose_name='Default amount')),
                ('cancellationPage', cms.models.fields.PageField(help_text='When the user returns to the site, send them to this page.', on_delete=django.db.models.deletion.CASCADE, related_name='cancellationPageFor', to='cms.Page', verbose_name='Cancellation Page')),
                ('successPage', cms.models.fields.PageField(help_text='When the user returns to the site after a successful transaction, send them to this page.', on_delete=django.db.models.deletion.CASCADE, related_name='successPageFor', to='cms.Page', verbose_name='Success Page')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.AddField(
            model_name='ipncartitem',
            name='ipn',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='paypal.IPNMessage'),
        ),
        migrations.AddField(
            model_name='ipncartitem',
            name='revenueItem',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='financial.RevenueItem'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-04-25 00:10
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import djangocms_text_ckeditor.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('cms', '0016_auto_20160608_1535'),
    ]

    operations = [
        migrations.CreateModel(
            name='FAQ',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('question', models.CharField(max_length=200)),
                ('answer', djangocms_text_ckeditor.fields.HTMLField(help_text='Answer the question.')),
                ('orderNum', models.PositiveIntegerField(default=0, help_text='This number specifies the order in which the questions will be shown on the FAQ page.')),
                ('draft', models.BooleanField(default=False, help_text='Check this box to prevent publication.')),
                ('creationDate', models.DateTimeField(auto_now_add=True)),
                ('modifiedDate', models.DateTimeField(auto_now=True)),
            ],
            options={
                'ordering': ('orderNum',),
            },
        ),
        migrations.CreateModel(
            name='FAQCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200, unique=True)),
                ('orderNum', models.PositiveIntegerField(default=0, help_text='This number specifies the order in which categories will be shown.')),
            ],
            options={
                'verbose_name_plural': 'FAQ Categories',
                'ordering': ('orderNum',),
            },
        ),
        migrations.CreateModel(
            name='FAQCategoryPluginModel',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='faq_faqcategorypluginmodel', serialize=False, to='cms.CMSPlugin')),
                ('showTitle', models.BooleanField(default=False, verbose_name='Show Category Title')),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='faq.FAQCategory')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='FAQSinglePluginModel',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='faq_faqsinglepluginmodel', serialize=False, to='cms.CMSPlugin')),
                ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='faq.FAQ')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.AddField(
            model_name='faq',
            name='category',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='faq.FAQCategory'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-08-09 02:47
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0020_auto_20180808_2247'),
        ('vouchers', '0003_auto_20170724_2103'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomerGroupVoucher',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.CustomerGroup', verbose_name='Customer group')),
                ('voucher', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='vouchers.Voucher', verbose_name='Voucher')),
            ],
            options={
                'verbose_name': 'Group-specific voucher restriction',
                'verbose_name_plural': 'Group-specific voucher restrictions',
            },
        ),
    ]



from django.dispatch import receiver
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from danceschool.core.signals import check_student_info
from danceschool.core.models import Customer
from danceschool.core.constants import getConstant

from .models import Requirement

import logging


# Define logger for this file
logger = logging.getLogger(__name__)


@receiver(check_student_info)
def checkRequirements(sender,**kwargs):
    '''
    Check that the customer meets all prerequisites for the items in the registration.
    '''

    if not getConstant('requirements__enableRequirements'):
        return

    logger.debug('Signal to check RegistrationContactForm handled by prerequisites app.')

    formData = kwargs.get('formData',{})
    first = formData.get('firstName')
    last = formData.get('lastName')
    email = formData.get('email')

    request = kwargs.get('request',{})
    registration = kwargs.get('registration',None)

    customer = Customer.objects.filter(
        first_name=first,
        last_name=last,
        email=email).first()

    requirement_warnings = []
    requirement_errors = []

    for ter in registration.temporaryeventregistration_set.all():
        if hasattr(ter.event,'getRequirements'):
            for req in ter.event.getRequirements():
                if not req.customerMeetsRequirement(
                    customer=customer,
                    danceRole=ter.role
                ):
                    if req.enforcementMethod == Requirement.EnforcementChoice.error:
                        requirement_errors.append((ter.event.name, req.name))
                    if req.enforcementMethod == Requirement.EnforcementChoice.warning:
                        requirement_warnings.append((ter.event.name,req.name))

    if requirement_errors:
        raise ValidationError(format_html(
            '<p>{}</p> <ul>{}</ul> <p>{}</p>',
            ugettext('Unfortunately, you do not meet the following requirements/prerequisites for the items you have chosen:\n'),
            mark_safe(''.join(['<li><em>%s:</em> %s</li>\n' % x for x in requirement_errors])),
            getConstant('requirements__errorMessage') or '',
        ))

    if requirement_warnings:
        messages.warning(request,format_html(
            '<p>{}</p> <ul>{}</ul> <p>{}</p>',
            mark_safe(ugettext('<strong>Please Note:</strong> It appears that you do not meet the following requirements/prerequisites for the items you have chosen:\n')),
            mark_safe(''.join(['<li><em>%s:</em> %s</li>\n' % x for x in requirement_warnings])),
            getConstant('requirements__warningMessage') or '',
        ))



from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool
from django.utils.translation import ugettext_lazy as _
from django.conf import settings


from .models import PayNowFormModel


class GiftCertificateFormPlugin(CMSPluginBase):
    model = PayNowFormModel
    name = _('Paypal Gift Certificate Form')
    render_template = "paypal/giftcertificate_form.html"
    cache = False
    module = 'Paypal'

    def render(self, context, instance, placeholder):
        ''' Create a UUID and check if a voucher with that ID exists before rendering '''
        context = super(GiftCertificateFormPlugin, self).render(context, instance, placeholder)

        # Paypal "live" mode is "production" to checkout.js
        mode = getattr(settings,'PAYPAL_MODE', 'sandbox')
        if mode == 'live':
            mode = 'production'

        context.update({
            'paypal_mode': mode,
        })

        return context


class CartPaymentFormPlugin(CMSPluginBase):
    model = PayNowFormModel
    name = _('Paypal Express Checkout Form')
    render_template = "paypal/express_checkout.html"
    cache = False
    module = 'Paypal'

    def render(self, context, instance, placeholder):
        ''' Add the cart-specific context to this form '''
        context = super(CartPaymentFormPlugin, self).render(context, instance, placeholder)

        # Paypal "live" mode is "production" to checkout.js
        mode = getattr(settings,'PAYPAL_MODE', 'sandbox')
        if mode == 'live':
            mode = 'production'

        context.update({
            'paypal_mode': mode,
        })

        return context


plugin_pool.register_plugin(GiftCertificateFormPlugin)
plugin_pool.register_plugin(CartPaymentFormPlugin)



# Give this app a custom verbose name to avoid confusion
from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class ThemeAppConfig(AppConfig):
    name = 'danceschool.themes'
    verbose_name = _('Theme Functions')



from django.core.management.base import BaseCommand
from django.apps import apps
from django.contrib.auth.models import Group, Permission

from six.moves import input

try:
    import readline
except ImportError:
    pass


class Command(BaseCommand):
    help = 'Create default groups and permissions for standard dance school setups'

    def boolean_input(self, question, default=None):
        '''
        Method for yes/no boolean inputs
        '''
        result = input("%s: " % question)
        if not result and default is not None:
            return default
        while len(result) < 1 or result[0].lower() not in "yn":
            result = input("Please answer yes or no: ")
        return result[0].lower() == "y"

    def handle(self, *args, **options):

        self.stdout.write(
            """
USER GROUPS AND PERMISSIONS
---------------------------

This project allows you to provide finely-grained permissions to individual users and
user groups, such as instructors and administrators.  This allows you to let different
types of users manage different types of content while still maintaining appropriate
security.

To get you started with the permissions system, we can create three initial user
groups, and give them different levels of permissions over content:

 - The "Board" group: Users in this group will receive permissions to edit
all public-facing content as well as all financial records.  They will not
automaticcaly receive permissions to edit certain other sitewide settings for
security reasons.
 - The "Instructor" group: Users in this group will receive permissions to use
school administrative functions such as emailing students, submitting expenses
and revenues, and viewing their own statistics and payment history.  However, by
default, these users cannot edit public-facing content such as page content or
FAQs.
 - The "Registration Desk" group: Users in this group receive only the ability
 to log into the site in order to view class registrations and check in students.
 By default, they cannot access any other administrative function.

We strongly encourage you to create these initial groups as a starting point for
managing staff permissions on the site.  The superuser that you created previously
will always retain permissions to edit all content and settings.  Additionally, you
can always go on to create additional groups, or to edit permissions on either a
group basis or an individual user basis.

Note: This process may take a minute or two to complete.

            """
        )

        create_board_group = self.boolean_input('Create \'Board\' group with default initial permissions [Y/n]', True)
        if create_board_group:
            board_group = Group.objects.get_or_create(name='Board')[0]

            # The Board group get all permissions on the CMS app and on all danceschool apps, plus
            # the permissions explicitly listed here by their natural_key.  Unfortunately this is
            # slow because we have to check permissions one-by-one
            give_explicit = [
                ('add_emailaddress', 'account', 'emailaddress'),
                ('change_emailaddress', 'account', 'emailaddress'),
                ('delete_emailaddress', 'account', 'emailaddress'),
                ('add_user', 'auth', 'user'),
                ('change_user', 'auth', 'user'),
            ]

            app_add_list = ['cms','core','djangocms_forms','djangocms_text_ckeditor','easy_thumbnails','filer']
            for this_app in [
                'danceschool.financial',
                'danceschool.discounts',
                'danceschool.faq',
                'danceschool.news',
                'danceschool.prerequisites',
                'danceschool.private_events',
                'danceschool.private_lessons',
                'danceschool.stats',
                'danceschool.vouchers',
                'danceschool.banlist',
                'danceschool.payments.paypal',
                'danceschool.payments.stripe',
            ]:
                if apps.is_installed(this_app):
                    app_add_list.append(this_app.split('.')[1])

            for perm in Permission.objects.all():
                if perm.natural_key() in give_explicit or perm.natural_key()[1] in app_add_list:
                    board_group.permissions.add(perm)
            self.stdout.write('Finished creating \'Board\' group and setting initial permissions.\n')

        create_instructor_group = self.boolean_input('Create \'Instructor\' group with default initial permissions [Y/n]', True)
        if create_instructor_group:
            instructor_group = Group.objects.get_or_create(name='Instructor')[0]

            give_explicit = [
                ('view_page', 'cms', 'page'),
                ('add_classdescription', 'core', 'classdescription'),
                ('change_classdescription', 'core', 'classdescription'),
                ('can_autocomplete_users', 'core', 'customer'),
                ('send_email', 'core', 'emailtemplate'),
                ('report_substitute_teaching', 'core', 'eventstaffmember'),
                ('update_instructor_bio', 'core', 'instructor'),
                ('view_own_instructor_finances', 'core', 'instructor'),
                ('view_own_instructor_stats', 'core', 'instructor'),
                ('process_refunds','core','invoice'),
                ('send_invoices', 'core', 'invoice'),
                ('view_all_invoices','core','invoice'),
                ('accept_door_payments', 'core', 'registration'),
                ('checkin_customers', 'core', 'registration'),
                ('override_register_closed', 'core', 'registration'),
                ('override_register_dropins', 'core', 'registration'),
                ('override_register_soldout', 'core', 'registration'),
                ('register_dropins', 'core', 'registration'),
                ('view_registration_summary', 'core', 'registration'),
                ('view_school_stats', 'core', 'staffmember'),
                ('view_staff_directory', 'core', 'staffmember'),
                ('add_file', 'filer', 'file'),
                ('change_file', 'filer', 'file'),
                ('can_use_directory_listing', 'filer', 'folder'),
                ('add_image', 'filer', 'image'),
                ('change_image', 'filer', 'image'),
                ('add_expenseitem', 'financial', 'expenseitem'),
                ('mark_expenses_paid', 'financial', 'expenseitem'),
                ('add_revenueitem', 'financial', 'revenueitem'),
                ('view_finances_bymonth', 'financial', 'revenueitem'),
                ('add_newsitem', 'news', 'newsitem'),
                ('change_newsitem', 'news', 'newsitem'),
                ('ignore_requirements', 'prerequisites', 'requirement'),
                ('add_eventreminder', 'private_events', 'eventreminder'),
                ('change_eventreminder', 'private_events', 'eventreminder'),
                ('delete_eventreminder', 'private_events', 'eventreminder'),
                ('add_privateevent', 'private_events', 'privateevent'),
                ('change_privateevent', 'private_events', 'privateevent'),
                ('delete_privateevent', 'private_events', 'privateevent'),
                ('edit_own_availability', 'private_lessons', 'instructoravailabilityslot'),
                ('view_banlist', 'banlist','bannedperson'),
            ]

            for perm in Permission.objects.all():
                if perm.natural_key() in give_explicit:
                    instructor_group.permissions.add(perm)
            self.stdout.write('Finished creating \'Instructor\' group and setting initial permissions.\n')

        create_regdesk_group = self.boolean_input('Create \'Registration Desk\' group with default initial permissions [Y/n]', True)
        if create_regdesk_group:
            regdesk_group = Group.objects.get_or_create(name='Registration Desk')[0]

            give_explicit = [
                ('view_page', 'cms', 'page'),
                ('can_autocomplete_users', 'core', 'customer'),
                ('process_refunds','core','invoice'),
                ('send_invoices', 'core', 'invoice'),
                ('view_all_invoices','core','invoice'),
                ('accept_door_payments', 'core', 'registration'),
                ('checkin_customers', 'core', 'registration'),
                ('override_register_closed', 'core', 'registration'),
                ('override_register_dropins', 'core', 'registration'),
                ('override_register_soldout', 'core', 'registration'),
                ('register_dropins', 'core', 'registration'),
                ('view_registration_summary', 'core', 'registration'),
                ('view_staff_directory', 'core', 'staffmember'),
                ('ignore_requirements', 'prerequisites', 'requirement'),
                ('view_banlist', 'banlist','bannedperson'),
            ]

            for perm in Permission.objects.all():
                if perm.natural_key() in give_explicit:
                    regdesk_group.permissions.add(perm)
            self.stdout.write('Finished creating \'Registration\' group and setting initial permissions.\n')



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-07-17 16:42
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('paypal', '0002_auto_20170614_1607'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='paymentrecord',
            options={'verbose_name': 'Paypal payment record', 'verbose_name_plural': 'Payment records'},
        ),
    ]



from django.db import models
from django.utils.translation import ugettext_lazy as _

from cms.models.pluginmodel import CMSPlugin


class StatsGraphPluginModel(CMSPlugin):
    '''
    A model for showing different types of school stats.  In general, the templates for these
    plugin instances call the specific URLs that they require to pull the data for the graph,
    so this model allows for only a template to be specified.
    '''
    template = models.CharField(verbose_name=_('Template'),max_length=250,null=True,blank=True)

    def get_short_description(self):
        desc = self.id
        choices = getattr(self.get_plugin_class(),'template_choices',[])
        choice_name = [x[1] for x in choices if x[0] == self.template]
        if choice_name:
            desc = choice_name[0]
        elif self.template:
            desc = self.template
        return desc



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-04-25 00:10
from __future__ import unicode_literals

import datetime
from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import filer.fields.file


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('filer', '0007_auto_20161016_1055'),
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExpenseCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Different types of tasks and payments should have different category names', max_length=50, unique=True)),
                ('defaultRate', models.FloatField(blank=True, help_text='This is the default hourly payment rate for this type of task.  For staff expenses and venue rentals, this will be overridden by the rate specified as default for the venue or staff type.', null=True, validators=[django.core.validators.MinValueValidator(0)])),
            ],
            options={
                'verbose_name_plural': 'Expense categories',
            },
        ),
        migrations.CreateModel(
            name='ExpenseItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('submissionDate', models.DateTimeField(default=datetime.datetime.now)),
                ('description', models.CharField(blank=True, max_length=200, null=True)),
                ('hours', models.FloatField(blank=True, help_text='Please indicate the number of hours to be paid for.', null=True, validators=[django.core.validators.MinValueValidator(0)])),
                ('wageRate', models.FloatField(blank=True, help_text='This should be filled automatically, but can be changed as needed.', null=True, validators=[django.core.validators.MinValueValidator(0)])),
                ('total', models.FloatField(blank=True, null=True, validators=[django.core.validators.MinValueValidator(0)])),
                ('adjustments', models.FloatField(default=0, help_text='Record any ex-post adjustments to the amount (e.g. refunds) in this field. A positive amount increases the netExpense, a negative amount reduces the netExpense.')),
                ('fees', models.FloatField(default=0, help_text='The sum of any transaction fees (e.g. Paypal fees) that were paid <strong>by us</strong>, and should therefore be added to net expense.')),
                ('paymentMethod', models.IntegerField(blank=True, choices=[(None, '-----'), (1, 'Cash'), (2, 'Paypal'), (3, 'Venmo'), (4, 'Bank/Debit Card'), (5, 'Check')], null=True, verbose_name='Payment Method')),
                ('comments', models.TextField(blank=True, null=True, verbose_name='Comments/Notes')),
                ('payToName', models.CharField(blank=True, max_length=30, null=True)),
                ('reimbursement', models.BooleanField(default=False, help_text='Check to indicate that this is a reimbursement expense (i.e. not compensation).')),
                ('approved', models.BooleanField(default=False, help_text='Check to indicate that expense is approved for payment.')),
                ('paid', models.BooleanField(default=False, help_text='Check to indicate that payment has been made.')),
                ('approvalDate', models.DateTimeField(blank=True, null=True)),
                ('paymentDate', models.DateTimeField(blank=True, null=True)),
                ('accrualDate', models.DateTimeField()),
            ],
            options={
                'permissions': (('mark_expenses_paid', 'Mark expenses as paid at the time of submission'),),
                'ordering': ['-accrualDate'],
            },
        ),
        migrations.CreateModel(
            name='RevenueCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Different types of revenue fall under different categories.', max_length=50, unique=True)),
                ('defaultAmount', models.FloatField(blank=True, help_text='This is the default amount of revenue for items in this category.', null=True, validators=[django.core.validators.MinValueValidator(0)])),
            ],
            options={
                'verbose_name_plural': 'Revenue categories',
            },
        ),
        migrations.CreateModel(
            name='RevenueItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('submissionDate', models.DateTimeField(default=datetime.datetime.now)),
                ('description', models.CharField(blank=True, max_length=200, null=True)),
                ('total', models.FloatField(help_text='The total revenue received, net of any discounts or voucher uses.  This is what we actually receive.', validators=[django.core.validators.MinValueValidator(0)])),
                ('grossTotal', models.FloatField(help_text='The gross total billed before the application of any discounts, or the use of any vouchers.', validators=[django.core.validators.MinValueValidator(0)], verbose_name='Gross Total')),
                ('adjustments', models.FloatField(default=0, help_text='Record any ex-post adjustments to the amount (e.g. refunds) in this field.  A positive amount increases the netRevenue, a negative amount reduces the netRevenue.')),
                ('fees', models.FloatField(default=0, help_text='The sum of any transaction fees (e.g. Paypal fees) that were paid <strong>by us</strong>, and should therefore be subtracted from net revenue.')),
                ('paymentMethod', models.IntegerField(choices=[(None, '-----'), (1, 'Cash'), (2, 'Paypal'), (3, 'Venmo'), (4, 'Bank/Debit Card'), (5, 'Check')], default=1, verbose_name='Payment Method')),
                ('invoiceNumber', models.CharField(blank=True, help_text='For Paypal payments, this will be the txn_id.  For cash payments, this will be automatically generated by the submission form.  More than one revenue item may have the same invoice number, because multiple events are paid for in one Paypal transaction.', max_length=80, null=True, verbose_name='Invoice Number')),
                ('comments', models.TextField(blank=True, null=True, verbose_name='Comments/Notes')),
                ('receivedFromName', models.CharField(blank=True, help_text='Enter who this revenue item was received from, if it is not associated with an existing registration.', max_length=50, null=True, verbose_name='Received From')),
                ('received', models.BooleanField(default=False, help_text='Check to indicate that payment has been received. Non-received payments are considered pending.')),
                ('receivedDate', models.DateTimeField(blank=True, null=True)),
                ('accrualDate', models.DateTimeField()),
                ('attachment', filer.fields.file.FilerFileField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='revenue_attachment', to='filer.File', verbose_name='Attach File (optional)')),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='financial.RevenueCategory')),
                ('currentlyHeldBy', models.ForeignKey(blank=True, help_text='If cash has not yet been deposited, this indicates who to contact in order to collect the cash for deposit.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='revenuesheldby', to=settings.AUTH_USER_MODEL, verbose_name='Cash currently in possession of')),
                ('event', models.ForeignKey(blank=True, help_text='If this item is associated with an Event, enter it here.', null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Event')),
                ('eventregistration', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.EventRegistration')),
            ],
            options={
                'permissions': (('export_financial_data', 'Export detailed financial transaction information to CSV'), ('view_finances_bymonth', 'View school finances month-by-month'), ('view_finances_byevent', 'View school finances by Event'), ('view_finances_detail', 'View school finances as detailed statement'), ('process_registration_refunds', 'Process registration refunds')),
                'ordering': ['-accrualDate'],
            },
        ),
    ]



from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from cms.toolbar_pool import toolbar_pool
from cms.toolbar_base import CMSToolbar
from cms.toolbar.items import BaseItem


@toolbar_pool.register
class PrivateLessonsToolbar(CMSToolbar):
    ''' Adds private events links to the Staff toolbar and category links to the related items toolbar '''

    def populate(self):
        staff_menu = self.toolbar.get_or_create_menu('core-staffmember',_('Staff'))
        position = staff_menu.find_first(BaseItem,name=_('Your Stats')) or 0

        staff_menu.add_link_item(_('Schedule a Private Lesson'), url=reverse('bookPrivateLesson'), position=position + 1)

        if (
            self.request.user.has_perm('private_lessons.edit_own_availability') or
            self.request.user.has_perm('private_lessons.edit_others_availability')
        ):
            staff_menu.add_link_item(_('Private Lesson Availability'), url=reverse('instructorAvailability'), position=position + 1)



from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool
from cms.models.pluginmodel import CMSPlugin

from django.utils.translation import ugettext_lazy as _

from .models import FAQSinglePluginModel, FAQCategoryPluginModel, FAQ


class FAQTOCPlugin(CMSPluginBase):
    model = CMSPlugin
    name = _('FAQ Table of Contents (auto-generated from FAQs on page)')
    render_template = 'faq/FAQ_TOC.html'
    cache = True
    module = _('FAQs')


class FAQAllPlugin(CMSPluginBase):
    model = CMSPlugin
    name = _('List of All FAQs')
    render_template = 'faq/FAQs.html'
    cache = True
    module = _('FAQs')

    def render(self, context, instance, placeholder):
        listing = FAQ.objects.filter(draft=False)

        context.update({
            'faq_list': listing,
        })
        return context


class FAQCategoryPlugin(CMSPluginBase):
    model = FAQCategoryPluginModel
    name = _('List of FAQs by Category')
    render_template = 'faq/FAQs.html'
    cache = True
    module = _('FAQs')

    def render(self, context, instance, placeholder):
        listing = FAQ.objects.filter(category=instance.category,draft=False)

        context.update({
            'faq_list': listing,
        })
        return context


class SingleQuestionPlugin(CMSPluginBase):
    model = FAQSinglePluginModel
    name = _('Individual FAQ Item')
    render_template = 'faq/single_FAQ.html'
    cache = True
    module = _('FAQs')

    def render(self, context, instance, placeholder):
        context.update({
            'faq_list': [instance.question,],
        })
        return context


plugin_pool.register_plugin(FAQTOCPlugin)
plugin_pool.register_plugin(FAQAllPlugin)
plugin_pool.register_plugin(FAQCategoryPlugin)
plugin_pool.register_plugin(SingleQuestionPlugin)



from django.dispatch import Signal

# Fires during the clean process of the StudentInfoView form, allowing hooked in apps to validate
# the form data and raise ValidationErrors or send warnings (messages) to the user by adding them to the
# request.
check_student_info = Signal(providing_args=['instance','formData','request','registration'])

# Fires after the student info form has been validated and the TemporaryRegistration record
# has been updated to reflect the submitted information from this form.  Since this signal is
# fired before price adjustments from vouchers and discounts are incorporated into the registration,
# this signal can be used to modify the TemporaryRegistration itself (be sure to save changes),
# and it can be used to make related changes in other apps (such as creating TemporaryVoucherUse)
# records in the VoucherUse app
post_student_info = Signal(providing_args=['registration',])

# Fires at the point when automatically-applied discounts may be applied to
# a preliminary registration.  Any handler that attaches to this signal should
# return an object that describes the discount (in the case of the discounts app,)
# a DiscountCombo object, as well as the discounted price to be applied to the _entire_
# cart, in tuple form as (object, discounted_price).
request_discounts = Signal(providing_args=['registration'])

# Fires after a discount has been actually applied, so that a hooked in discounts
# app can make a record of the discount having been applied.  Note that the core
# app by default records the discounted price of a registration, net of all discounts
# and also of all voucher uses, but it does not itself record information on the
# discounts or vouchers actually applied.
apply_discount = Signal(providing_args=['registration', 'discount','discount_amount'])

# Fires at the point when free add-on items may be applied to
# a preliminary registration.  Add-ons are simply descriptive, they do not link
# to a registration or any other object at present.
# Any handler that attaches to this signal should return a list with the names of
# the addons.
apply_addons = Signal(providing_args=['registration'])

# Fires when vouchers or any other direct price adjustments are ready to be applied.
# Any handler that attaches to this signal should return a name/description of the
# adjustment as well as the amount of the adjustment, in tuple form as (name,amount).
apply_price_adjustments = Signal(providing_args=['registration','initial_price'])

# Fires after a Registration is created.
post_registration = Signal(providing_args=['registration',])

# Fires on the customer profile page to collect customer information from other apps
# without overriding the CustomerStatsView.
get_customer_data = Signal(providing_args=['customer',])



# Python Imports
import logging
import string
import random

# Third Party Imports
from django.dispatch import receiver
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.translation import ugettext
from django.utils.translation import ugettext_lazy as _
from django.db.models import Q

# Local Application Specific Imports
from danceschool.banlist.models import BannedPerson, BanFlaggedRecord
from danceschool.core.signals import check_student_info
from danceschool.core.constants import getConstant, REG_VALIDATION_STR
from danceschool.core.tasks import sendEmail

# Define logger for this file
logger = logging.getLogger(__name__)


def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


@receiver(check_student_info)
def checkBanlist(sender, **kwargs):
    '''
    Check that this individual is not on the ban list.
    '''

    if not getConstant('registration__enableBanList'):
        return

    logger.debug('Signal to check RegistrationContactForm handled by banlist app.')

    formData = kwargs.get('formData', {})
    first = formData.get('firstName')
    last = formData.get('lastName')
    email = formData.get('email')

    request = kwargs.get('request', {})
    session = getattr(request, 'session', {}).get(REG_VALIDATION_STR, {})
    registrationId = getattr(kwargs.get('registration', None), 'id', None)

    records = BannedPerson.objects.exclude(
        disabled=True
    ).exclude(
        expirationDate__lte=timezone.now()
    ).filter(
        (Q(firstName__iexact=first) & Q(lastName__iexact=last)) |
        Q(bannedemail__email__iexact=email)
    )

    if not records.exists():
        return

    # Generate an "error code" to reference so that it is easier to lookup
    # the record on why they were flagged.
    flagCode = ''.join(random.choice(string.ascii_uppercase) for x in range(8))
    ip = get_client_ip(request)
    respondTo = getConstant('registration__banListContactEmail') or getConstant('contact__businessEmail')

    for record in records:
        flagRecord = BanFlaggedRecord.objects.create(
            flagCode=flagCode,
            person=record,
            ipAddress=ip,
            data={'session': session, 'formData': formData, 'registrationId': registrationId}
        )

    notify = getConstant('registration__banListNotificationEmail')
    if notify:
        send_from = getConstant('contact__businessEmail')
        subject = _('Notice of attempted registration by banned individual')
        message = _(
            'This is an automated notification that the following individual has attempted ' +
            'to register for a class series or event:\n\n' +
            'Name: %s\n' % record.fullName +
            'Email: %s\n' % email +
            'Date/Time: %s\n' % flagRecord.dateTime +
            'IP Address: %s\n\n' % ip +
            'This individual has been prevented from finalizing their registration, and they ' +
            'have been asked to notify the school at %s with code %s to proceed.' % (respondTo, flagCode)
        )

        sendEmail(subject, message, send_from, to=[notify])

    message = ugettext('There appears to be an issue with this registration. '
                       'Please contact %s to proceed with the registration process. '
                       'You may reference the error code %s.' % (respondTo, flagCode))

    if request.user.has_perm('banlist.ignore_ban'):
        messages.warning(request, message)

    else:
        raise ValidationError(message)



import logging

from django.core.mail import send_mail

from .constants import getConstant


# Define logger for this file
logger = logging.getLogger(__name__)


def emailErrorMessage(subject,message):
    '''
    Useful for sending error messages via email.
    '''
    if not getConstant('email__enableErrorEmails'):
        logger.info('Not sending error email: error emails are not enabled.')
        return

    send_from = getConstant('email__errorEmailFrom')
    send_to = getConstant('email__errorEmailTo')

    if not send_from or not send_to:
        logger.error('Cannot send error emails because addresses have not been specified.')
        return

    try:
        send_mail(subject,message,
                  send_from,
                  [send_to], fail_silently=False)
        logger.debug('Error email sent.')
    except Exception as e:
        logger.error('Error email was not sent: %s' % e)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-08-09 02:47
from __future__ import unicode_literals

import danceschool.core.mixins
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0019_invoice_buyerpayssalestax'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomerGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, verbose_name='Group name')),
            ],
            options={
                'ordering': ('name',),
                'verbose_name': 'Customer group',
                'verbose_name_plural': 'Customer groups',
            },
            bases=(danceschool.core.mixins.EmailRecipientMixin, models.Model),
        ),
        migrations.AddField(
            model_name='customer',
            name='groups',
            field=models.ManyToManyField(blank=True, help_text='Customer groups may be used for group-specific discounts and vouchers, as well as for email purposes.', to='core.CustomerGroup', verbose_name='Customer groups'),
        ),
    ]



from django.http import HttpResponseRedirect, Http404, HttpResponseBadRequest
from django.shortcuts import get_object_or_404, get_list_or_404
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.views.generic import FormView, CreateView, UpdateView, DetailView, TemplateView, ListView
from django.db.models import Min, Q, Count
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone
from django.contrib.auth.models import User
from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.contrib.auth.mixins import AccessMixin
from django.contrib.contenttypes.models import ContentType

from calendar import month_name
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from urllib.parse import unquote_plus, unquote
from braces.views import UserFormKwargsMixin, PermissionRequiredMixin, LoginRequiredMixin, StaffuserRequiredMixin
from cms.constants import RIGHT
from cms.models import Page
import re
import logging

from .models import Event, Series, PublicEvent, EventOccurrence, EventRole, EventRegistration, StaffMember, Instructor, Invoice, Customer
from .forms import SubstituteReportingForm, InstructorBioChangeForm, RefundForm, EmailContactForm, RepeatEventForm, InvoiceNotificationForm
from .constants import getConstant, EMAIL_VALIDATION_STR, REFUND_VALIDATION_STR
from .mixins import EmailRecipientMixin, StaffMemberObjectMixin, FinancialContextMixin, AdminSuccessURLMixin, EventOrderMixin
from .signals import get_customer_data
from .utils.requests import getIntFromGet


# Define logger for this file
logger = logging.getLogger(__name__)


class EventRegistrationSelectView(PermissionRequiredMixin, EventOrderMixin, ListView):
    '''
    This view is used to select an event for viewing registration data in the EventRegistrationSummaryView
    '''
    template_name = 'core/events_viewregistration_list.html'
    permission_required = 'core.view_registration_summary'
    reverse_time_ordering = True

    def get_queryset(self):
        return Event.objects.filter(
            Q(startTime__gte=timezone.now() - timedelta(days=90)) & (
                Q(series__isnull=False) | Q(publicevent__isnull=False)
            )
        ).annotate(count=Count('eventregistration')).annotate(**self.get_annotations()).exclude(
            Q(count=0) & Q(status__in=[Event.RegStatus.hidden, Event.RegStatus.regHidden, Event.RegStatus.disabled])
        )


class EventRegistrationSummaryView(PermissionRequiredMixin, DetailView):
    '''
    This view is used to access the set of registrations for a given series or event
    '''
    template_name = 'core/view_eventregistrations.html'
    permission_required = 'core.view_registration_summary'

    def get_object(self, queryset=None):
        return get_object_or_404(
            Event.objects.filter(id=self.kwargs.get('series_id')))

    def get_context_data(self,**kwargs):
        ''' Add the list of registrations for the given series '''
        context = {
            'event': self.object,
            'registrations': EventRegistration.objects.filter(
                event=self.object,
                cancelled=False
            ).order_by('registration__customer__user__first_name','registration__customer__user__last_name'),
        }
        context.update(kwargs)
        return super(EventRegistrationSummaryView,self).get_context_data(**context)


#################################
# Used for various form submission redirects (called by the AdminSuccessURLMixin)


class SubmissionRedirectView(TemplateView):
    template_name = 'cms/forms/submission_redirect.html'

    def get_context_data(self, **kwargs):
        context = super(SubmissionRedirectView,self).get_context_data(**kwargs)

        try:
            redirect_url = unquote(self.request.GET.get('redirect_url',''))
            if not redirect_url:
                redirect_url = Page.objects.get(pk=getConstant('general__defaultAdminSuccessPage')).get_absolute_url(settings.LANGUAGE_CODE)
        except ObjectDoesNotExist:
            redirect_url = '/'

        context.update({
            'redirect_url': redirect_url,
            'seconds': self.request.GET.get('seconds',5),
        })

        return context

################################################
# For Viewing Invoices and sending notifications


class ViewInvoiceView(AccessMixin, FinancialContextMixin, DetailView):
    template_name = 'core/invoice.html'
    model = Invoice

    def get(self,request,*args,**kwargs):
        '''
        Invoices can be viewed only if the validation string is provided, unless
        the user is logged in and has view_all_invoice permissions
        '''
        self.object = self.get_object()
        if request.GET.get('v',None) == self.object.validationString or request.user.has_perm('core.view_all_invoices'):
            return super(ViewInvoiceView,self).get(request,*args,**kwargs)
        return self.handle_no_permission()

    def get_context_data(self, **kwargs):
        context = super(ViewInvoiceView,self).get_context_data(**kwargs)
        context.update({
            'invoice': self.object,
            'payments': self.get_payments(),
        })
        return context

    def get_payments(self):
        if not getattr(self,'payments',None):
            self.payments = self.object.get_payments()
        return self.payments


class InvoiceNotificationView(FinancialContextMixin,AdminSuccessURLMixin, PermissionRequiredMixin, StaffuserRequiredMixin, FormView):
    success_message = _('Invoice notifications successfully sent.')
    template_name = 'core/invoice_notification.html'
    permission_required = 'core.send_invoices'
    form_class = InvoiceNotificationForm

    def form_valid(self, form):
        invoice_ids = [
            k.replace('invoice_','') for k,v in form.cleaned_data.items() if 'invoice_' in k and v is True
        ]
        invoices = [x for x in self.toNotify if str(x.id) in invoice_ids]

        for invoice in invoices:
            if invoice.get_default_recipients():
                invoice.sendNotification()

        messages.success(self.request, self.success_message)
        return HttpResponseRedirect(self.get_success_url())

    def dispatch(self, request, *args, **kwargs):
        ''' Get the set of invoices for which to permit notifications '''

        if 'pk' in self.kwargs:
            try:
                self.invoices = Invoice.objects.filter(pk=self.kwargs.get('pk'))[:]
            except ValueError:
                raise Http404()
            if not self.invoices:
                raise Http404()
        else:
            ids = request.GET.get('invoices','')
            try:
                self.invoices = Invoice.objects.filter(id__in=[x for x in ids.split(',')])[:]
            except ValueError:
                return HttpResponseBadRequest(_('Invalid invoice identifiers specified.'))

        if not self.invoices:
            return HttpResponseBadRequest(_('No invoice identifiers specified.'))

        toNotify = []
        cannotNotify = []

        for invoice in self.invoices:
            if invoice.get_default_recipients():
                toNotify.append(invoice)
            else:
                cannotNotify.append(invoice)
        self.toNotify = toNotify
        self.cannotNotify = cannotNotify

        return super(InvoiceNotificationView, self).dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        ''' Pass the set of invoices to the form for creation '''
        kwargs = super(InvoiceNotificationView,self).get_form_kwargs()
        kwargs['invoices'] = self.toNotify
        return kwargs

    def get_context_data(self,**kwargs):
        context = super(InvoiceNotificationView,self).get_context_data(**kwargs)
        context.update({
            'toNotify': self.toNotify,
            'cannotNotify': self.cannotNotify,
        })
        return context


#################################
# Refund processing and confirmation step views


class RefundConfirmationView(FinancialContextMixin, AdminSuccessURLMixin, PermissionRequiredMixin, StaffuserRequiredMixin, SuccessMessageMixin, TemplateView):
    success_message = _('Refund successfully processed.')
    template_name = 'core/refund_confirmation.html'
    permission_required = 'core.process_refunds'

    def get(self, request, *args, **kwargs):
        self.form_data = request.session.get(REFUND_VALIDATION_STR,{}).get('form_data',{})
        if not self.form_data:
            return HttpResponseRedirect(reverse('refundProcessing', args=(self.form_data.get('id'),)))

        try:
            self.invoice = Invoice.objects.get(id=self.form_data.get('id'))
        except ObjectDoesNotExist:
            return HttpResponseRedirect(reverse('refundProcessing', args=(self.form_data.get('id'),)))

        self.payments = self.invoice.get_payments()

        if request.GET.get('confirmed','').lower() == 'true' and self.payments:
            return self.process_refund()
        return super(RefundConfirmationView,self).get(request, *args, **kwargs)

    def get_context_data(self,**kwargs):
        context = super(RefundConfirmationView,self).get_context_data(**kwargs)

        total_refund_amount = self.form_data['total_refund_amount']
        initial_refund_amount = self.form_data['initial_refund_amount']
        amount_to_refund = max(total_refund_amount - initial_refund_amount,0)

        context.update({
            'form_data': self.form_data,
            'payments': self.payments,
            'total_refund_amount': total_refund_amount,
            'initial_refund_amount': initial_refund_amount,
            'amount_to_refund': amount_to_refund,
        })
        return context

    def process_refund(self):

        refund_data = self.invoice.data.get('refunds',[])

        total_refund_amount = self.form_data['total_refund_amount']
        initial_refund_amount = self.form_data['initial_refund_amount']
        amount_to_refund = max(total_refund_amount - initial_refund_amount,0)

        # Keep track of total refund fees as well as how much reamins to be
        # refunded as we iterate through payments to refund them.
        remains_to_refund = amount_to_refund

        for this_payment in self.payments:
            if remains_to_refund <= 0:
                break
            if not this_payment.refundable:
                continue

            this_payment_amount = this_payment.netAmountPaid or 0
            this_refund_amount = min(this_payment_amount, remains_to_refund)

            # This dictionary will be updated and then added to refund_data for
            # this invoice whether the refund is successful or not
            this_refund_response_data = {
                'datetime': timezone.now(),
                'id': this_payment.recordId,
                'methodName': this_payment.methodName,
                'amount': this_refund_amount,
            }

            this_refund_response = this_payment.refund(this_refund_amount)

            if not this_refund_response:
                # If no response is received, then we must stop because we cannot be sure that a refund has
                # not already been processed.
                this_refund_response_data.update({
                    'status': 'error',
                    'errorMessage': str(_('Error: No response from payment processing app. Check payment processor records for refund status.')),
                })
            elif this_refund_response[0].get('status').lower() == 'success':
                # A successful refund returns {'status': 'success'}
                amount_refunded = this_refund_response[0].get('refundAmount',0)
                fees = this_refund_response[0].get('fees',0)

                this_refund_response_data.update({
                    'status': 'success',
                    'refundAmount': amount_refunded,
                    'fees': fees,
                    'response': [dict(this_refund_response[0]),],
                })

                # Incrementally update the invoice's refunded amount
                self.invoice.adjustments -= amount_refunded
                self.invoice.amountPaid -= amount_refunded
                self.invoice.fees += fees
                remains_to_refund -= amount_refunded

            else:
                this_refund_response_data.update({
                    'status': 'error',
                    'errorMessage': _('An unkown error has occurred. Check payment processor records for refund status.'),
                    'response': [dict(this_refund_response[0]),],
                    'id': this_payment.recordId,
                    'methodName': this_payment.methodName,
                    'invoice': self.invoice.id,
                    'refundAmount': this_refund_amount,
                })
            refund_data.append(this_refund_response_data)

            if this_refund_response_data.get('status') == 'error':
                logger.error(this_refund_response_data.get('errorMessage'))
                logger.error(this_refund_response_data)

                messages.error(self.request, this_refund_response_data.get('errorMessage'))

                self.invoice.data['refunds'] = refund_data
                self.invoice.save()
                self.invoice.allocateFees()
                self.request.session.pop(REFUND_VALIDATION_STR,None)
                return HttpResponseRedirect(self.get_success_url())

        # If there were no errors, then check to ensure that the entire request refund was refunded.
        # If so, then return success, otherwise return indicating that the refund was not completely
        # applied.
        if abs(remains_to_refund) <= 0.01:
            messages.success(self.request, self.success_message)
        else:
            messages.error(self.request, _('Error, not all of the requested refund was applied. Check invoice and payment processor records for details.'))

        self.invoice.data['refunds'] = refund_data
        if self.invoice.total + self.invoice.adjustments == 0:
            self.invoice.status = Invoice.PaymentStatus.fullRefund
        self.invoice.save()

        # Identify the items to which refunds should be allocated and update the adjustments line
        # for those items.  Fees are also allocated across the items for which the refund was requested.
        item_refund_data = [(k.split('_')[2],v) for k,v in self.form_data.items() if k.startswith('item_refundamount_')]
        for item_request in item_refund_data:
            this_item = self.invoice.invoiceitem_set.get(id=item_request[0])
            this_item.adjustments = -1 * float(item_request[1])
            this_item.save()

            add_taxes = this_item.taxes if self.invoice.buyerPaysSalesTax else 0

            # If the refund is a complete refund, then cancel the EventRegistration entirely.
            if abs(this_item.total + add_taxes + this_item.adjustments) < 0.01 and this_item.finalEventRegistration:
                this_item.finalEventRegistration.cancelled = True
                this_item.finalEventRegistration.save()

        # Ensure that all fees are allocated appropriately
        self.invoice.allocateFees()

        self.request.session.pop(REFUND_VALIDATION_STR,None)
        return HttpResponseRedirect(self.get_success_url())


class RefundProcessingView(FinancialContextMixin, PermissionRequiredMixin, StaffuserRequiredMixin, UpdateView):
    template_name = 'core/process_refund.html'
    form_class = RefundForm
    permission_required = 'core.process_refunds'
    model = Invoice

    def get_context_data(self,**kwargs):
        context = super(RefundProcessingView,self).get_context_data(**kwargs)
        context.update({
            'invoice': self.object,
            'payments': self.get_payments(),
        })
        if self.object.finalRegistration:
            context['registration'] = self.object.finalRegistration
        elif self.object.temporaryRegistration:
            context['registration'] = self.object.temporaryRegistration

        return context

    def form_valid(self, form):
        # Avoid JSON serialization issues by passing the Invoice ID, not the object itself
        clean_data = form.cleaned_data
        clean_data['id'] = str(clean_data.get('id').id)

        self.request.session[REFUND_VALIDATION_STR] = {
            'form_data': clean_data,
            'invoice': str(self.object.id),
        }
        return HttpResponseRedirect(reverse('refundConfirmation'))

    def get_payments(self):
        if not getattr(self,'payments',None):
            self.payments = self.object.get_payments()
        return self.payments


#################################
# Email view function and form


class EmailConfirmationView(AdminSuccessURLMixin, PermissionRequiredMixin, TemplateView):
    permission_required = 'core.send_email'
    template_name = 'core/email_confirmation_page.html'
    success_message = _('Email sent successfully.')

    def get(self, request, *args, **kwargs):
        self.form_data = request.session.get(EMAIL_VALIDATION_STR,{}).get('form_data',{})
        if not self.form_data:
            return HttpResponseRedirect(reverse('emailStudents'))
        if request.GET.get('confirmed','').lower() == 'true':
            return self.send_email()
        return super(EmailConfirmationView,self).get(request, *args, **kwargs)

    def send_email(self):
        subject = self.form_data.pop('subject')
        message = self.form_data.pop('message')
        html_message = self.form_data.pop('html_message',None)
        richTextChoice = self.form_data.pop('richTextChoice')
        cc_myself = self.form_data.pop('cc_myself')
        testemail = self.form_data.pop('testemail')
        month = self.form_data.pop('month')
        series = self.form_data.pop('series')
        customers = self.form_data.pop('customers',[])

        email_kwargs = {
            'from_name': self.form_data['from_name'],
            'from_address': self.form_data['from_address'],
        }

        if richTextChoice == 'HTML':
            email_kwargs.update({
                'send_html': True,
                'html_message': html_message,
            })

        items_to_send = []
        if month is not None and month != '':
            items_to_send += list(Series.objects.filter(month=datetime.strptime(month,'%m-%Y').month, year=datetime.strptime(month,'%m-%Y').year))
        if series not in [None,'',[],['']]:
            items_to_send += list(Event.objects.filter(id__in=series))
        if customers:
            items_to_send.append(list(Customer.objects.filter(id__in=customers)))

        # We always call one email per series so that the series-level tags
        # can be passed.  The entire list of customers is also a single item
        # in the items_to_send list, because they can be processed all at once.
        for s in items_to_send:
            if isinstance(s,Event):
                regs = EventRegistration.objects.filter(event=s,cancelled=False)
                emails = []
                for x in regs:
                    emails += x.get_default_recipients() or []
            else:
                # Customers are themselves the list.
                regs = s
                emails = [x.email for x in s]

            email_kwargs['cc'] = []
            if cc_myself:
                email_kwargs['cc'].append(email_kwargs['from_address'])

            email_kwargs['bcc'] = [email_kwargs['from_address'] or getConstant('email__defaultEmailFrom'),]

            if testemail:
                message = str(_('Test email from %s to be sent to: ' % email_kwargs['from_address'])) + '\n\n'
                message += ', '.join(email_kwargs['bcc']) + ', '.join(emails) + '\n\n'
                message += str(_('Email body:')) + '\n\n' + message
                email_kwargs['bcc'] = []

            # If there are no context tags, then this can be sent as a single bulk email.
            # Otherwise, send a separate email for each event registration
            has_tags = re.search('\{\{.+\}\}',message)
            if not has_tags:
                email_kwargs['bcc'] += emails
                # Avoid duplicate emails
                email_kwargs['bcc'] = list(set(email_kwargs['bcc']))

                # instantiate the recipient mixin directly
                email_class = EmailRecipientMixin()
                email_class.email_recipient(subject,message,**email_kwargs)
            else:
                for r in regs:
                    r.email_recipient(subject,message,**email_kwargs)

        self.request.session.pop(EMAIL_VALIDATION_STR,None)
        messages.success(self.request, self.success_message)

        return HttpResponseRedirect(self.get_success_url())

    def get_context_data(self,**kwargs):
        context = super(EmailConfirmationView,self).get_context_data(**kwargs)
        context.update(self.form_data)

        month = self.form_data['month']
        series = self.form_data['series']
        customers = self.form_data.get('customers')
        from_address = self.form_data['from_address']
        cc_myself = self.form_data['cc_myself']

        events_to_send = []
        if month is not None and month != '':
            events_to_send += Series.objects.filter(month=datetime.strptime(month,'%m-%Y').month, year=datetime.strptime(month,'%m-%Y').year)
        if series not in [None,'',[],['']]:
            events_to_send += [Event.objects.get(id=x) for x in series]

        # We always call one email per series so that the series-level tags
        # can be passed.
        regs = EventRegistration.objects.filter(event__in=events_to_send)

        customerSet = Customer.objects.filter(id__in=customers) if customers else []

        emails = [r.customer.email for r in regs] + [r.email for r in customerSet]
        cc = []
        if cc_myself:
            cc.append(from_address)
        bcc = [getConstant('email__defaultEmailFrom')]

        context.update({
            'events_to_send': events_to_send,
            'customers_to_send': customerSet,
            'emails': emails,
            'cc': cc,
            'bcc': bcc,
        })

        return context


class SendEmailView(PermissionRequiredMixin, UserFormKwargsMixin, FormView):
    form_class = EmailContactForm
    permission_required = 'core.send_email'
    template_name = 'cms/forms/display_form_classbased_admin.html'

    def dispatch(self, request, *args, **kwargs):
        ''' If a list of customers or groups was passed, then parse it '''
        ids = request.GET.get('customers')
        groups = request.GET.get('customergroup')
        self.customers = None

        if ids or groups:
            # Initial filter applies to no one but allows appending by logical or
            filters = Q(id__isnull=True)

            if ids:
                filters = filters | Q(id__in=[int(x) for x in ids.split(',')])
            if groups:
                filters = filters | Q(groups__id__in=[int(x) for x in groups.split(',')])

            try:
                self.customers = Customer.objects.filter(filters)
            except ValueError:
                return HttpResponseBadRequest(_('Invalid customer ids passed'))

        return super(SendEmailView,self).dispatch(request, *args, **kwargs)

    def get_form_kwargs(self, **kwargs):
        '''
        Get the list of recent months and recent series to pass to the form
        '''

        numMonths = 12
        lastStart = Event.objects.annotate(Min('eventoccurrence__startTime')).order_by('-eventoccurrence__startTime__min').values_list('eventoccurrence__startTime__min',flat=True).first()
        if lastStart:
            month = lastStart.month
            year = lastStart.year
        else:
            month = timezone.now().month
            year = timezone.now().year

        months = [('',_('None'))]
        for i in range(0,numMonths):
            newmonth = (month - i - 1) % 12 + 1
            newyear = year
            if month - i - 1 < 0:
                newyear = year - 1
            newdate = datetime(year=newyear,month=newmonth,day=1)
            newdateStr = newdate.strftime("%m-%Y")
            monthStr = newdate.strftime("%B, %Y")
            months.append((newdateStr,monthStr))

        cutoff = timezone.now() - timedelta(days=120)

        allEvents = Event.objects.filter(startTime__gte=cutoff).order_by('-startTime')

        kwargs = super(SendEmailView, self).get_form_kwargs(**kwargs)
        kwargs.update({
            "months": months,
            "recentseries": [('','None')] + [(x.id,'%s %s: %s' % (month_name[x.month],x.year,x.name)) for x in allEvents],
            "customers": self.customers,
        })
        return kwargs

    def get_initial(self):
        '''
        If the user already submitted the form and decided to return from the
        confirmation page, then re-populate the form
        '''
        initial = super(SendEmailView,self).get_initial()

        form_data = self.request.session.get(EMAIL_VALIDATION_STR,{}).get('form_data',{})
        if form_data:
            initial.update(form_data)
        return initial

    def get_context_data(self,**kwargs):
        context = super(SendEmailView,self).get_context_data(**kwargs)

        context.update({
            'form_title': _('Email Students'),
            'form_description': _('Use this form to contact current or recent students.'),
        })
        return context

    def form_valid(self, form):
        ''' Pass form data to the confirmation view '''
        form.cleaned_data.pop('template',None)
        self.request.session[EMAIL_VALIDATION_STR] = {'form_data': form.cleaned_data}
        return HttpResponseRedirect(reverse('emailConfirmation'))


############################################
# Customer and Instructor Stats Views


class AccountProfileView(LoginRequiredMixin, DetailView):
    model = User
    template_name = 'core/account_profile.html'

    def get_object(self, queryset=None):
        return self.request.user

    def get_context_data(self,**kwargs):
        context = {}
        user = self.get_object()

        context.update({
            'primary_email': user.emailaddress_set.filter(primary=True).first(),
            'verified_emails': user.emailaddress_set.filter(verified=True),
            'unverified_emails': user.emailaddress_set.filter(verified=False),
        })

        if hasattr(user,'customer'):
            context.update({
                'customer': user.customer,
                'customer_verified': user.emailaddress_set.filter(email=user.customer.email,verified=True).exists(),
            })
            context['customer_eventregs'] = EventRegistration.objects.filter(registration__customer=user.customer)

        context['verified_eventregs'] = EventRegistration.objects.filter(
            registration__customer__email__in=[x.email for x in context['verified_emails']]
        ).exclude(
            id__in=[x.id for x in context.get('customer_eventregs',[])]
        )
        context['submitted_eventregs'] = EventRegistration.objects.filter(
            registration__invoice__submissionUser=self.request.user,registration__payAtDoor=False
        ).exclude(
            id__in=[x.id for x in context.get('customer_eventregs',[])]
        ).exclude(
            id__in=[x.id for x in context.get('verified_eventregs',[])]
        )

        if hasattr(user,'staffmember'):
            context.update({
                'staffmember': user.staffmember,
                'upcoming_events': Event.objects.filter(endTime__gt=timezone.now(),eventstaffmember__staffMember=user.staffmember).distinct().order_by('-startTime'),
            })

        # Get any extra context data passed by other apps.  These data require unique keys, so when writing
        # a handler for this signal, be sure to provide unique context keys.
        if hasattr(user,'customer'):
            extra_customer_data = get_customer_data.send(
                sender=AccountProfileView,
                customer=user.customer,
            )
            for item in extra_customer_data:
                if len(item) > 1 and isinstance(item[1],dict):
                    # Ensure that 'customer' is not overwritten and add everything else
                    item[1].pop('customer',None)
                    context.update(item[1])

        return super(AccountProfileView,self).get_context_data(**context)


class OtherAccountProfileView(PermissionRequiredMixin, AccountProfileView):
    permission_required = 'core.view_other_user_profiles'

    def get_object(self, queryset=None):
        if 'user_id' in self.kwargs:
            return get_object_or_404(User.objects.filter(id=self.kwargs.get('user_id')))
        else:
            return self.request.user


class InstructorStatsView(StaffMemberObjectMixin, PermissionRequiredMixin, DetailView):
    model = StaffMember
    template_name = 'core/instructor_stats.html'
    permission_required = 'core.view_own_instructor_stats'

    def get_context_data(self,**kwargs):
        instructor = self.object
        context = {}

        context.update({
            'instructor': instructor,
            'prior_series': Event.objects.filter(startTime__lte=timezone.now(),eventstaffmember__staffMember=instructor).order_by('-startTime'),
            'upcoming_series': Event.objects.filter(startTime__gt=timezone.now(),eventstaffmember__staffMember=instructor).order_by('-startTime'),
        })

        if context['prior_series']:
            context.update({'first_series': context['prior_series'].last(),})
            context.update({
                'teaching_since': month_name[context['first_series'].month] + ' ' + str(context['first_series'].year),
                'student_count': sum([x.numRegistered for x in context['prior_series']]),
            })
        context.update({'series_count': len(context['prior_series']) + len(context['upcoming_series'])})

        # Note: This get the detailview's context, not all the mixins.  Supering itself led to an infinite loop.
        return super(DetailView, self).get_context_data(**context)


class OtherInstructorStatsView(InstructorStatsView):
    permission_required = 'core.view_other_instructor_stats'

    def get_object(self, queryset=None):
        if 'first_name' in self.kwargs and 'last_name' in self.kwargs:
            return get_object_or_404(
                StaffMember.objects.filter(**{'firstName': unquote_plus(self.kwargs['first_name']).replace('_',' '), 'lastName': unquote_plus(self.kwargs['last_name']).replace('_',' ')}))
        else:
            return None


#####################################
# Individual Class Series/Event Views


class IndividualClassView(FinancialContextMixin, TemplateView):
    template_name = 'core/individual_class.html'

    def get(self,request,*args,**kwargs):
        # These are passed via the URL
        year = self.kwargs.get('year')
        month = self.kwargs.get('month')
        session_slug = self.kwargs.get('session_slug')
        slug = self.kwargs.get('slug','')

        if month:
            try:
                month_number = list(month_name).index(month or 0)
            except ValueError:
                raise Http404(_('Invalid month.'))

        filters = ~Q(status=Event.RegStatus.hidden) \
            & ~Q(status=Event.RegStatus.linkOnly) \
            & Q(classDescription__slug=slug)
        if year and month:
            filters = filters & Q(year=year or None) & Q(month=month_number or None)
        if session_slug:
            filters = filters & Q(session__slug=session_slug)

        seriesset = get_list_or_404(Series,filters)

        # This will pass through to the context data by default
        kwargs.update({'seriesset': seriesset})

        # For each Series in the set, add a button to the toolbar to edit the Series details
        if hasattr(request,'user') and request.user.has_perm('core.change_series'):
            for this_series in seriesset:
                this_title = _('Edit Class Details')
                if len(seriesset) > 1:
                    this_title += ' (#%s)' % this_series.id
                request.toolbar.add_button(this_title, reverse('admin:core_series_change', args=([this_series.id,])), side=RIGHT)

        return super(IndividualClassView,self).get(request,*args,**kwargs)


class IndividualEventView(FinancialContextMixin, TemplateView):
    template_name = 'core/individual_event.html'

    def get(self,request,*args,**kwargs):
        # These are passed via the URL
        year = self.kwargs.get('year',timezone.now().year)
        month = self.kwargs.get('month',0)
        session_slug = self.kwargs.get('session_slug')
        slug = self.kwargs.get('slug','')

        if month:
            try:
                month_number = list(month_name).index(month or 0)
            except ValueError:
                raise Http404(_('Invalid month.'))

        filters = ~Q(status=Event.RegStatus.hidden) \
            & ~Q(status=Event.RegStatus.linkOnly) \
            & Q(slug=slug)
        if year and month:
            filters = filters & Q(year=year or None) & Q(month=month_number or None)
        if session_slug:
            filters = filters & Q(session__slug=session_slug)

        eventset = get_list_or_404(PublicEvent,filters)

        # If an alternative link is given by one or more of these events, then redirect to that.
        overrideLinks = [x.link for x in eventset if x.link]
        if overrideLinks:
            return HttpResponseRedirect(overrideLinks[0])

        # This will pass through to the context data by default
        kwargs.update({'eventset': eventset})

        # For each Event in the set, add a button to the toolbar to edit the Event details
        if hasattr(request,'user') and request.user.has_perm('core.change_publicevent'):
            for this_event in eventset:
                this_title = _('Edit Event Details')
                if len(eventset) > 1:
                    this_title += ' (#%s)' % this_event.id
                request.toolbar.add_button(this_title, reverse('admin:core_publicevent_change', args=([this_event.id,])), side=RIGHT)

        return super(IndividualEventView,self).get(request,*args,**kwargs)


#####################################
# View for Repeating Events from admin
class RepeatEventsView(SuccessMessageMixin, AdminSuccessURLMixin, PermissionRequiredMixin, FormView):
    '''
    This view is for an admin action to repeat events.
    '''
    template_name = 'core/repeat_events.html'
    form_class = RepeatEventForm
    permission_required = 'core.add_event'
    success_message = _('Repeated events created successfully.')

    def dispatch(self, request, *args, **kwargs):
        ids = request.GET.get('ids')
        ct = getIntFromGet(request,'ct')

        try:
            contentType = ContentType.objects.get(id=ct)
            self.objectClass = contentType.model_class()
        except (ValueError, ObjectDoesNotExist):
            return HttpResponseBadRequest(_('Invalid content type passed.'))

        # This view only deals with subclasses of Events (Public Events, Series, etc.)
        if not isinstance(self.objectClass(),Event):
            return HttpResponseBadRequest(_('Invalid content type passed.'))

        try:
            self.queryset = self.objectClass.objects.filter(id__in=[int(x) for x in ids.split(',')])
        except ValueError:
            return HttpResponseBadRequest(_('Invalid ids passed'))

        return super(RepeatEventsView,self).dispatch(request,*args,**kwargs)

    def get_context_data(self,**kwargs):
        context = super(RepeatEventsView,self).get_context_data(**kwargs)
        context.update({
            'events': self.queryset,
        })

        return context

    def form_valid(self, form):
        ''' For each object in the queryset, create the duplicated objects '''

        startDate = form.cleaned_data.get('startDate')
        repeatEvery = form.cleaned_data.get('repeatEvery')
        periodicity = form.cleaned_data.get('periodicity')
        quantity = form.cleaned_data.get('quantity')
        endDate = form.cleaned_data.get('endDate')

        # Create a list of start dates, based on the passed  values of repeatEvery,
        # periodicity, quantity and endDate.  This list will be iterated through to
        # create the new instances for each event.
        if periodicity == 'D':
            delta = {'days': repeatEvery}
        elif periodicity == 'W':
            delta = {'weeks': repeatEvery}
        elif periodicity == 'M':
            delta = {'months': repeatEvery}

        repeat_list = []
        this_date = startDate

        if quantity:
            for k in range(0,quantity):
                repeat_list.append(this_date)
                this_date = this_date + relativedelta(**delta)
        elif endDate:
            while (this_date <= endDate):
                repeat_list.append(this_date)
                this_date = this_date + relativedelta(**delta)

        # Now, loop through the events in the queryset to create duplicates of them
        for event in self.queryset:

            # For each new occurrence, we determine the new startime by the distance from
            # midnight of the first occurrence date, where the first occurrence date is
            # replaced by the date given in repeat list
            old_min_time = event.startTime.replace(hour=0,minute=0,second=0,microsecond=0)

            old_occurrence_data = [
                (x.startTime - old_min_time, x.endTime - old_min_time, x.cancelled)
                for x in event.eventoccurrence_set.all()
            ]

            old_role_data = [(x.role, x.capacity) for x in event.eventrole_set.all()]

            for instance_date in repeat_list:

                # Ensure that time zones are treated properly
                new_datetime = timezone.now().replace(
                    year=instance_date.year,month=instance_date.month,day=instance_date.day,
                    hour=0,minute=0,second=0,microsecond=0
                )

                # Removing the pk and ID allow new instances of the event to
                # be created upon saving with automatically generated ids.
                event.id = None
                event.pk = None
                event.save()

                # Create new occurrences
                for occurrence in old_occurrence_data:
                    EventOccurrence.objects.create(
                        event=event,
                        startTime=new_datetime + occurrence[0],
                        endTime=new_datetime + occurrence[1],
                        cancelled=occurrence[2],
                    )

                # Create new event-specific role data
                for role in old_role_data:
                    EventRole.objects.create(
                        event=event,
                        role=role[0],
                        capacity=role[1],
                    )

                # Need to save twice to ensure that startTime etc. get
                # updated properly.
                event.save()

            return super(RepeatEventsView,self).form_valid(form)


############################################################
# View for instructors to report that they substitute taught
#

class SubstituteReportingView(AdminSuccessURLMixin, PermissionRequiredMixin, UserFormKwargsMixin, SuccessMessageMixin, CreateView):
    '''
    This view is used to report substitute teaching.
    '''
    template_name = 'cms/forms/display_form_classbased_admin.html'
    form_class = SubstituteReportingForm
    permission_required = 'core.report_substitute_teaching'
    success_message = _('Substitute teaching reported successfully.')

    def get_context_data(self,**kwargs):
        context = super(SubstituteReportingView,self).get_context_data(**kwargs)

        context.update({
            'form_title': _('Report Substitute Teaching'),
            'form_description': _('Use this form to report substitute teaching.'),
        })
        return context


############################################################
# View for instructors to change their bio information
#


class InstructorBioChangeView(AdminSuccessURLMixin, StaffMemberObjectMixin, PermissionRequiredMixin, SuccessMessageMixin, UpdateView):
    '''
    This view now permits changing the instructor's bio information.
    '''
    model = Instructor
    template_name = 'cms/forms/display_form_classbased_admin.html'
    form_class = InstructorBioChangeForm
    permission_required = 'core.update_instructor_bio'
    success_message = _('Instructor information updated successfully.')

    def get_context_data(self,**kwargs):
        context = super(InstructorBioChangeView,self).get_context_data(**kwargs)

        context.update({
            'form_title': _('Update Contact Information'),
            'form_description': _('Use this form to update your contact information.'),
        })
        return context


############################################################
# View for Instructor/Staff directory
#


class StaffDirectoryView(PermissionRequiredMixin, ListView):
    '''
    This view shows a directory of instructors/staff
    '''
    template_name = 'core/staff_directory.html'
    permission_required = 'core.view_staff_directory'
    queryset = StaffMember.objects.exclude(instructor__status__in=[
        Instructor.InstructorStatus.retired,
        Instructor.InstructorStatus.retiredGuest,
        Instructor.InstructorStatus.hidden,
    ])



from django.utils.translation import ugettext_lazy as _

from djangocms_picture.cms_plugins import PicturePlugin
from cms.plugin_pool import plugin_pool

from danceschool.core.mixins import PluginTemplateMixin
from danceschool.core.registries import plugin_templates_registry, PluginTemplateBase

class PictureSplashTemplatePlugin(PicturePlugin):
    '''
    A subclass of the Django CMS PicturePlugin that can be used just to select
    a photo for inclusion as the splash page in a theme.  Your theme will need
    to override 'djangocms_picture/splash_image/picture.html' to place the
    splash image appropriately in the template (e.g. using the <header> tag).
    '''

    name = _('Template Splash Image')

    fieldsets = [
        (None, {
            'fields': (
                'picture',
                'external_picture',
                'attributes',
            )
        }),
    ]

    def get_render_template(self, context, instance, placeholder):
        ''' Template cannot be chosen when using this plugin. '''
        return 'djangocms_picture/splash_image/picture.html'

plugin_pool.register_plugin(PictureSplashTemplatePlugin)



'''
This file defines a variety of preferences that must be set in the DB,
but can be changed dynamically.
'''

from django.utils.translation import ugettext_lazy as _
from django.template.loader import get_template

from dynamic_preferences.types import BooleanPreference, IntegerPreference, ModelChoicePreference, Section
from dynamic_preferences.registries import global_preferences_registry

from danceschool.core.models import EventStaffCategory, EmailTemplate, get_defaultEmailName, get_defaultEmailFrom

# we create some section objects to link related preferences together

privateLessons = Section('privateLessons', _('Private Lessons'))


############################
# Private Lessons Preferences
#
@global_preferences_registry.register
class AllowPublicBooking(BooleanPreference):
    section = privateLessons
    name = 'allowPublicBooking'
    verbose_name = _('Allow non-staff to book private lessons for themselves.')
    help_text = _('The private lesson booking functions will only be available to individuals with permissions unless this box is checked.')
    default = True


@global_preferences_registry.register
class AllowRegistration(BooleanPreference):
    section = privateLessons
    name = 'allowRegistration'
    verbose_name = _('Allow payment for private lessons through the registration system.')
    help_text = _('If the studio handles payment for private lessons, then checking this box allows users to process payment through the regular online registration system.  If the studio only handles booking for private lessons, but not payment, then this box should not be checked.')
    default = True


@global_preferences_registry.register
class NotifyPrivateLessonInstructor(BooleanPreference):
    section = privateLessons
    name = 'notifyInstructor'
    verbose_name = _('Notify instructor when a private lesson is booked')
    help_text = _('If checked, then instructors will receive a notification email whenever a private lesson is scheduled with them.')
    default = True


@global_preferences_registry.register
class OpenBookingDaysAhead(IntegerPreference):
    section = privateLessons
    name = 'openBookingDays'
    verbose_name = _('Open private lesson booking days in advance')
    help_text = _('If instructor availability has been configured, permitted users will be able to book private lessons beginning this many days in advance.  Enter 0 for no restiction.')
    default = 30


@global_preferences_registry.register
class CloseBookingDaysAhead(IntegerPreference):
    section = privateLessons
    name = 'closeBookingDays'
    verbose_name = _('Close private lesson booking days in advance')
    help_text = _('Enter 0 to allow booking up to the start time of the lesson.')
    default = 0


@global_preferences_registry.register
class DefaultLessonLength(IntegerPreference):
    section = privateLessons
    name = 'defaultLessonLength'
    verbose_name = _('Default lesson booking length (minutes)')
    default = 60


@global_preferences_registry.register
class LessonLengthInterval(IntegerPreference):
    section = privateLessons
    name = 'lessonLengthInterval'
    verbose_name = _('Interval in which lesson lengths can be changed (minutes)')
    default = 30


@global_preferences_registry.register
class MinimumLessonLength(IntegerPreference):
    section = privateLessons
    name = 'minimumLessonLength'
    verbose_name = _('Minimum lesson booking length (minutes)')
    help_text = _('Enter 0 for no limit')
    default = 30


@global_preferences_registry.register
class MaximumLessonLength(IntegerPreference):
    section = privateLessons
    name = 'maximumLessonLength'
    verbose_name = _('Maximum lesson booking length (minutes)')
    help_text = _('Enter 0 for no limit')
    default = 90


@global_preferences_registry.register
class StaffCategoryPrivateLesson(ModelChoicePreference):
    section = privateLessons
    name = 'eventStaffCategoryPrivateLesson'
    verbose_name = _('Private Lesson Event Staff Category')
    model = EventStaffCategory
    queryset = EventStaffCategory.objects.all()

    def get_default(self):
        return EventStaffCategory.objects.get_or_create(name=_('Private Lesson Instruction'))[0]


@global_preferences_registry.register
class LessonBookedEmailTemplate(ModelChoicePreference):
    section = privateLessons
    name = 'lessonBookedEmailTemplate'
    verbose_name = _('Email template used to notify customers that their lesson is scheduled')
    help_text = _('This email template will only be used for lesson bookings that do not go through the full online registration and payment system.')
    model = EmailTemplate
    queryset = EmailTemplate.objects.all()

    def get_default(self):

        initial_template = get_template('email/private_lesson_registration_success.html')
        with open(initial_template.origin.name,'r') as infile:
            content = infile.read()
            infile.close()

        return EmailTemplate.objects.get_or_create(
            name=_('Private Lesson Booking Confirmation Email'),
            defaults={
                'subject': _('Confirmation of Scheduled Private Lesson'),
                'content': content or '',
                'defaultFromAddress': get_defaultEmailFrom(),
                'defaultFromName': get_defaultEmailName(),
                'defaultCC': '',
                'hideFromForm': True,
            }
        )[0]


@global_preferences_registry.register
class LessonBookedInstructorEmailTemplate(ModelChoicePreference):
    section = privateLessons
    name = 'lessonBookedInstructorEmailTemplate'
    verbose_name = _('Email template used to notify instructors that a lesson has been scheduled')
    model = EmailTemplate
    queryset = EmailTemplate.objects.all()

    def get_default(self):

        initial_template = get_template('email/private_lesson_booking_alert.html')
        with open(initial_template.origin.name,'r') as infile:
            content = infile.read()
            infile.close()

        return EmailTemplate.objects.get_or_create(
            name=_('Private Lesson Booking Instructor Notification'),
            defaults={
                'subject': _('Notification of Private Lesson Scheduling'),
                'content': content or '',
                'defaultFromAddress': get_defaultEmailFrom(),
                'defaultFromName': get_defaultEmailName(),
                'defaultCC': '',
                'hideFromForm': True,
            }
        )[0]



from django.core.mail import send_mail
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from huey import crontab
from huey.contrib.djhuey import db_periodic_task
import logging

from danceschool.core.constants import getConstant
from .models import EventReminder


# Define logger for this file
logger = logging.getLogger(__name__)


@db_periodic_task(crontab(minute='*'))
def sendReminderEmails():

    if not getConstant('general__enableCronTasks'):
        return

    reminders_needed = EventReminder.objects.filter(**{
        'time__lte': timezone.now(),
        'completed':False,
        'notifyList__isnull': False})
    if reminders_needed:
        for note in reminders_needed:
            for user in note.notifyList.all():
                sent = sendReminderEmailToUser(user,note)
                if sent:
                    # Mark reminder as sent so it won't be sent twice.
                    note.completed = True
                    note.save()
                    logger.info("Email notification sent to user: " + user.first_name + ' ' + user.last_name + ' at ' + user.email)
                else:
                    logger.warning("Unable to send email to user: " + user.first_name + ' ' + user.last_name + ' at ' + user.email)
    else:
        logger.debug("No notifications to send!")
        pass


def sendReminderEmailToUser(user,reminder):
    subject = _('REMINDER: ') + reminder.eventOccurrence.event.name
    subject += ' on ' + reminder.eventOccurrence.startTime.strftime('%A, %B %-d, %Y')
    if not reminder.eventOccurrence.allDay:
        subject += ' at ' + reminder.eventOccurrence.startTime.strftime('%-I:%M %p')

    content = render_to_string('private_events/reminder_emails.html',{
        'name': ' '.join([user.first_name, user.last_name]),
        'event': reminder.eventOccurrence.event,
        'occurrence': reminder.eventOccurrence,
        'businessName': getConstant('contact__businessName'),
    })

    try:
        if getConstant('email__disableSiteEmails') and getConstant('email__enableErrorEmails'):
            sent = send_mail(
                subject,content,
                getConstant('email__defaultEmailFrom'),
                [
                    getConstant('email__errorEmailTo'),
                ],
                fail_silently=False)
        else:
            sent = send_mail(
                subject,content,
                getConstant('email__defaultEmailFrom'),
                [
                    user.email,
                ],
                fail_silently=False)
    except Exception as e:
        logger.error('Error in sending reminder emails: %s' % e)
        return False
    if sent:
        return True
    return False
